import os.path

import nltk
from PySide6.QtCore import QObject, QThread, Slot, Signal
from sqlalchemy import func

from airunner.data.models.settings_models import AIModels, ControlnetModel
from airunner.data.bootstrap.controlnet_bootstrap_data import controlnet_bootstrap_data
from airunner.data.bootstrap.sd_file_bootstrap_data import SD_FILE_BOOTSTRAP_DATA
from airunner.data.bootstrap.llm_file_bootstrap_data import LLM_FILE_BOOTSTRAP_DATA
from airunner.data.bootstrap.whisper import WHISPER_FILES
from airunner.data.bootstrap.speech_t5 import SPEECH_T5_FILES
from airunner.enums import SignalCode
from airunner.mediator_mixin import MediatorMixin
from airunner.settings import NLTK_DOWNLOAD_DIR
from airunner.utils.network.huggingface_downloader import HuggingfaceDownloader
from airunner.windows.main.settings_mixin import SettingsMixin
from airunner.windows.setup_wizard.base_wizard import BaseWizard
from airunner.windows.setup_wizard.installation_settings.templates.install_page_ui import Ui_install_page
from airunner.settings import DEFAULT_PATH_SETTINGS
from airunner.utils.os.create_airunner_directory import create_airunner_paths

nltk.data.path.append(NLTK_DOWNLOAD_DIR)

CONTROLNET_PATHS = [
    "lllyasviel/control_v11p_sd15_canny",
    "lllyasviel/control_v11f1p_sd15_depth",
    "lllyasviel/control_v11f1p_sd15_depth",
    "lllyasviel/control_v11f1p_sd15_depth",
    "lllyasviel/control_v11f1p_sd15_depth",
    "lllyasviel/control_v11p_sd15_mlsd",
    "lllyasviel/control_v11p_sd15_normalbae",
    "lllyasviel/control_v11p_sd15_normalbae",
    "lllyasviel/control_v11p_sd15_scribble",
    "lllyasviel/control_v11p_sd15_scribble",
    "lllyasviel/control_v11p_sd15_seg",
    "lllyasviel/control_v11p_sd15_lineart",
    "lllyasviel/control_v11p_sd15_lineart",
    "lllyasviel/control_v11p_sd15s2_lineart_anime",
    "lllyasviel/control_v11p_sd15_openpose",
    "lllyasviel/control_v11p_sd15_openpose",
    "lllyasviel/control_v11p_sd15_openpose",
    "lllyasviel/control_v11p_sd15_openpose",
    "lllyasviel/control_v11p_sd15_openpose",
    "lllyasviel/control_v11p_sd15_softedge",
    "lllyasviel/control_v11p_sd15_softedge",
    "lllyasviel/control_v11p_sd15_softedge",
    "lllyasviel/control_v11p_sd15_softedge",
    "lllyasviel/control_v11e_sd15_ip2p",
    "lllyasviel/control_v11p_sd15_inpaint",
    "lllyasviel/control_v11e_sd15_shuffle",
]


class InstallWorker(
    QObject,
    MediatorMixin,
    SettingsMixin
):
    file_download_finished = Signal()
    progress_updated = Signal(int, int)

    def __init__(self, parent):
        MediatorMixin.__init__(self)
        
        super(InstallWorker, self).__init__()
        self.parent = parent
        self.current_step = -1
        self.total_models_in_current_step = 0
        self.hf_downloader = HuggingfaceDownloader(
            lambda a, b: self.progress_updated.emit(a, b)
        )
        self.hf_downloader.completed.connect(
            lambda: self.file_download_finished.emit()
        )
        self.register(SignalCode.DOWNLOAD_COMPLETE, self.download_finished)
        self.register(SignalCode.PATH_SET, self.path_set)

    def download_stable_diffusion(self):
        self.parent.on_set_downloading_status_label({
            "label": "Downloading Stable Diffusion models..."
        })

        
        models = self.session.query(AIModels).filter(
            AIModels.category == "stablediffusion",
            AIModels.is_default == 1,
            AIModels.version != "SDXL Turbo"
        ).all()
        

        self.total_models_in_current_step += len(models)
        for model in models:
            if model.name == "CompVis Safety Checker":
                action_key = "safety_checker"
                action = f"{model.pipeline_action}/{action_key}"
            elif model.name == "OpenAI Feature Extractor":
                action_key = "feature_extractor"
                action = f"{model.pipeline_action}/{action_key}"
            else:
                action = model.pipeline_action
                action_key = model.pipeline_action
            files = SD_FILE_BOOTSTRAP_DATA[model.version][action_key]
            self.parent.total_steps += len(files)
            for filename in files:
                requested_file_path = os.path.expanduser(
                    os.path.join(
                        self.path_settings.base_path,
                        model.model_type,
                        "models",
                        model.version,
                        action
                    )
                )
                try:
                    self.hf_downloader.download_model(
                        requested_path=model.path,
                        requested_file_name=filename,
                        requested_file_path=requested_file_path,
                        requested_callback=self.progress_updated.emit
                    )
                except Exception as e:
                    print(f"Error downloading {filename}: {e}")

    def download_controlnet(self):
        self.parent.on_set_downloading_status_label({
            "label": "Downloading Controlnet models..."
        })
        self.total_models_in_current_step += len(controlnet_bootstrap_data)
        for controlnet_model in controlnet_bootstrap_data:
            files = SD_FILE_BOOTSTRAP_DATA[controlnet_model["version"]]["controlnet"]
            self.parent.total_steps += len(files)
            for filename in files:
                requested_file_path = os.path.expanduser(
                    os.path.join(
                        self.path_settings.base_path,
                        "art",
                        "models",
                        controlnet_model["version"],
                        "controlnet",
                        controlnet_model["path"]
                    )
                )
                try:
                    self.hf_downloader.download_model(
                        requested_path=controlnet_model["path"],
                        requested_file_name=filename,
                        requested_file_path=requested_file_path,
                        requested_callback=self.progress_updated.emit
                    )
                except Exception as e:
                    print(f"Error downloading {filename}: {e}")

    def download_controlnet_processors(self):
        self.parent.on_set_downloading_status_label({
            "label": "Downloading Controlnet processors..."
        })
        controlnet_processor_files = [
            "150_16_swin_l_oneformer_coco_100ep.pth",
            "250_16_swin_l_oneformer_ade20k_160k.pth",
            "ControlNetHED.pth",
            "ControlNetLama.pth",
            "RealESRGAN_x4plus.pth",
            "ZoeD_M12_N.pt",
            "body_pose_model.pth",
            "clip_g.pth",
            "dpt_hybrid-midas-501f0c75.pt",
            "erika.pth",
            "facenet.pth",
            "hand_pose_model.pth",
            "lama.ckpt",
            "latest_net_G.pth",
            "mlsd_large_512_fp32.pth",
            "netG.pth",
            "network-bsds500.pth",
            "res101.pth",
            "scannet.pt",
            "sk_model.pth",
            "sk_model2.pth",
            "table5_pidinet.pth",
            "upernet_global_small.pth",
        ]
        self.parent.total_steps += len(controlnet_processor_files)
        for filename in controlnet_processor_files:
            requested_file_path = os.path.expanduser(
                os.path.join(
                    self.path_settings.base_path,
                    "art",
                    "models",
                    "controlnet_processors"
                )
            )
            try:
                self.hf_downloader.download_model(
                    requested_path=f"lllyasviel/Annotators",
                    requested_file_name=filename,
                    requested_file_path=requested_file_path,
                    requested_callback=self.progress_updated.emit
                )
            except Exception as e:
                print(f"Error downloading {filename}: {e}")

    def download_llms(self):
        
        models = self.session.query(AIModels).filter(
            AIModels.category == "llm",
            AIModels.is_default == 1
        ).all()
        
        self.total_models_in_current_step += len(models)
        for model in models:
            files = LLM_FILE_BOOTSTRAP_DATA[model.path]["files"]
            self.parent.total_steps += len(files)
            for filename in files:
                requested_file_path = os.path.expanduser(
                    os.path.join(
                        self.path_settings.base_path,
                        "text",
                        "models",
                        model.category,
                        model.pipeline_action,
                        model.path
                    )
                )
                try:
                    self.hf_downloader.download_model(
                        requested_path=model.path,
                        requested_file_name=filename,
                        requested_file_path=requested_file_path,
                        requested_callback=self.progress_updated.emit
                    )
                except Exception as e:
                    print(f"Error downloading {filename}: {e}")

    def download_stt(self):
        self.parent.on_set_downloading_status_label({
            "label": "Downloading STT models..."
        })
        for k, v in WHISPER_FILES.items():
            self.total_models_in_current_step += len(v)
        for k, v in WHISPER_FILES.items():
            for filename in v:
                requested_file_path = os.path.expanduser(
                    os.path.join(
                        self.path_settings.base_path,
                        "text",
                        "models",
                        "stt",
                        k
                    )
                )
                try:
                    self.hf_downloader.download_model(
                        requested_path=k,
                        requested_file_name=filename,
                        requested_file_path=requested_file_path,
                        requested_callback=self.progress_updated.emit
                    )
                except Exception as e:
                    print(f"Error downloading {filename}: {e}")

    def download_tts(self):
        self.parent.on_set_downloading_status_label({
            "label": "Downloading TTS models..."
        })
        for k, v in SPEECH_T5_FILES.items():
            self.total_models_in_current_step += len(v)
            for filename in v:
                requested_file_path = os.path.expanduser(
                    os.path.join(
                        self.path_settings.base_path,
                        "text",
                        "models",
                        "tts",
                        k
                    )
                )
                try:
                    self.hf_downloader.download_model(
                        requested_path=k,
                        requested_file_name=filename,
                        requested_file_path=requested_file_path,
                        requested_callback=self.progress_updated.emit
                    )
                except Exception as e:
                    print(f"Error downloading {filename}: {e}")

    def download_nltk_files(self):
        self.parent.on_set_downloading_status_label({
            "label": "Downloading NLTK models..."
        })
        path = os.path.expanduser(os.path.join(
            NLTK_DOWNLOAD_DIR,
            "corpora"
        ))
        os.makedirs(path, exist_ok=True)
        path = os.path.expanduser(os.path.join(
            NLTK_DOWNLOAD_DIR,
            "tokenizers",
            "punkt"
        ))
        os.makedirs(path, exist_ok=True)
        nltk.download(
            "stopwords",
            download_dir=NLTK_DOWNLOAD_DIR,
            quiet=True,
            halt_on_error=False,
            raise_on_error=False
        )
        nltk.download(
            "punkt",
            download_dir=NLTK_DOWNLOAD_DIR,
            quiet=True,
            halt_on_error=False,
            raise_on_error=False
        )
        self.download_finished()

    def finalize_installation(self, *args):
        self.parent.on_set_downloading_status_label({
            "label": "Installation complete."
        })
        self.parent.update_progress_bar()
        self.parent.parent.show_final_page()

    def update_progress(self, current, total):
        print("update progress", current, total)

    @Slot()
    def download_finished(self):
        self.total_models_in_current_step -= 1
        if self.total_models_in_current_step <= 0:
            self.set_page()
    
    @Slot()
    def path_set(self):
        self.set_page()

    def run(self):
        if (
            self.application_settings.user_agreement_checked and
            self.application_settings.stable_diffusion_agreement_checked and
            self.application_settings.airunner_agreement_checked
        ):
            self.current_step = -1
            self.set_page()

    def set_page(self):
        if (
            self.application_settings.stable_diffusion_agreement_checked and
            self.current_step == -1
        ):
            """
            Create the airunner paths
            """
            create_airunner_paths(self.path_settings)
            self.update_application_settings("paths_initialized", True)
            self.parent.on_set_downloading_status_label({
                "label": f"Downloading Stable Diffusion"
            })
            self.current_step = 1
            self.download_stable_diffusion()
        elif (
            self.application_settings.stable_diffusion_agreement_checked and
            self.current_step == 1
        ):
            self.parent.on_set_downloading_status_label({
                "label": f"Downloading Controlnet"
            })
            self.current_step = 2
            self.download_controlnet()
        elif (
            self.application_settings.stable_diffusion_agreement_checked and
            self.current_step == 2
        ):
            self.current_step = 3
            self.download_controlnet_processors()
        elif self.current_step == 3:
            self.parent.on_set_downloading_status_label({
                "label": f"Downloading LLM"
            })
            self.current_step = 4
            self.download_llms()
        elif self.current_step == 4:
            self.parent.on_set_downloading_status_label({
                "label": f"Downloading Text-to-Speech"
            })
            self.current_step = 5
            self.download_tts()
        elif self.current_step == 5:
            self.parent.on_set_downloading_status_label({
                "label": f"Downloading NLTK files"
            })
            self.current_step = 6
            self.download_nltk_files()
        elif self.current_step == 6:
            self.parent.on_set_downloading_status_label({
                "label": f"Downloading Speech-to-Text"
            })
            self.current_step = 7
            self.download_stt()
        elif self.current_step == 7:
             self.hf_downloader.download_model(
                 requested_path="",
                 requested_file_name="",
                 requested_file_path="",
                 requested_callback=self.finalize_installation
             )


class InstallPage(BaseWizard):
    class_name_ = Ui_install_page

    def __init__(self, parent):
        super(InstallPage, self).__init__(parent)
        self.steps_completed = 0

        # reset the progress bar
        self.ui.progress_bar.setValue(0)
        self.ui.progress_bar.setMaximum(100)

        # These will increase
        self.total_steps = 0
        if self.application_settings.stable_diffusion_agreement_checked:
            self.total_steps += 1

        

        controlnet_model_count = self.session.query(func.count(ControlnetModel.id.distinct())).scalar()
        controlnet_version_count = self.session.query(func.count(ControlnetModel.version.distinct())).scalar()

        llm_model_count = self.session.query(func.count(AIModels.id)).filter(AIModels.category == 'llm').scalar()

        self.total_steps += controlnet_model_count * controlnet_version_count
        self.total_steps += llm_model_count
        self.total_steps += len(SPEECH_T5_FILES["microsoft/speecht5_tts"])
        self.total_steps += len(WHISPER_FILES["openai/whisper-tiny"])

        self.register(SignalCode.DOWNLOAD_COMPLETE, self.update_progress_bar)
        self.register(SignalCode.DOWNLOAD_PROGRESS, self.download_progress)
        self.register(SignalCode.UPDATE_DOWNLOAD_LOG, self.update_download_log)
        self.register(SignalCode.CLEAR_DOWNLOAD_STATUS_BAR, self.clear_status_bar)
        self.register(SignalCode.SET_DOWNLOAD_STATUS_LABEL, self.on_set_downloading_status_label)

        self.thread = QThread()
        self.worker = InstallWorker(self)
        self.worker.moveToThread(self.thread)
    
    def start(self):
        self.thread.started.connect(self.worker.run)
        self.thread.start()

    def on_set_downloading_status_label(self, data: dict = None):
        if "message" in data:
            self.set_status(data["message"])

        if "label" in data:
            self.ui.status_bar.setFormat(data["label"])

    def download_progress(self, data: dict):
        if data["current"] == 0:
            progress = 0
        else:
            progress = data["current"] / data["total"]
        self.ui.status_bar.setValue(progress * 100)

    def update_progress_bar(self):
        self.steps_completed += 1
        if self.total_steps == self.steps_completed:
            self.ui.progress_bar.setValue(100)
        else:
            self.ui.progress_bar.setValue((self.steps_completed / self.total_steps) * 100)

    def set_status(self, message: str):
        # set the text of a QProgressBar
        self.ui.status_bar.setFormat(message)

    def update_download_log(self, data: dict):
        self.ui.log.appendPlainText(data["message"]+"\n")

    def clear_status_bar(self):
        self.ui.status.setText("")
        self.ui.status_bar.setValue(0)



from airunner.settings import DEFAULT_LLM_HF_PATH
from airunner.windows.download_wizard.download_thread import DownloadThread
from airunner.windows.download_wizard.download_wizard_page import DownloadWizardPage
from airunner.windows.setup_wizard.model_setup.llm.templates.llm_setup_ui import Ui_llm_setup


class LLMSetup(DownloadWizardPage):
    class_name_ = Ui_llm_setup

    def start_download(self):
        self.models_to_download = [
            {
                "model": {
                    "path": DEFAULT_LLM_HF_PATH
                },
            }
        ]
        self.download_thread = DownloadThread(self.models_to_download)
        self.download_thread.progress_updated.connect(self.update_progress)
        self.download_thread.download_finished.connect(self.download_finished)
        self.download_thread.start()



import logging
import datetime
import os
from typing import List, Type

from sqlalchemy import create_engine
from sqlalchemy.orm import joinedload, sessionmaker, scoped_session

from airunner.data.models.settings_models import Chatbot, AIModels, Schedulers, Lora, PathSettings, SavedPrompt, \
    Embedding, PromptTemplate, ControlnetModel, FontSetting, PipelineModel, ShortcutKeys, \
    GeneratorSettings, WindowSettings, ApplicationSettings, ActiveGridSettings, ControlnetSettings, \
    ImageToImageSettings, OutpaintSettings, DrawingPadSettings, MetadataSettings, \
    LLMGeneratorSettings, TTSSettings, SpeechT5Settings, EspeakSettings, STTSettings, BrushSettings, GridSettings, \
    MemorySettings, Message, Conversation, Summary, ImageFilterValue, TargetFiles, WhisperSettings, Base, User
from airunner.enums import SignalCode
from airunner.utils.image.convert_binary_to_image import convert_binary_to_image


class SettingsMixinSharedInstance:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(SettingsMixinSharedInstance, cls).__new__(cls, *args, **kwargs)
            cls._instance._initialized = False
        return cls._instance

    def __init__(self):
        if self._initialized:
            return
        self.db_path = os.path.expanduser(
            os.path.join(
                "~",
                ".local",
                "share",
                "airunner",
                "data",
                "airunner.db"
            )
        )
        self.engine = create_engine(f'sqlite:///{self.db_path}')
        Base.metadata.create_all(self.engine)
        self.Session = scoped_session(sessionmaker(bind=self.engine))
        self.conversation_id = None

        # Configure the logger
        self.logger = logging.getLogger("AI Runner")
        self.logger.setLevel(logging.DEBUG)

        # Remove all existing handlers
        if self.logger.hasHandlers():
            self.logger.handlers.clear()

        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(filename)s - %(funcName)s - %(lineno)d - %(message)s'))
        self.logger.addHandler(handler)

        # Disable propagation to the root logger
        self.logger.propagate = False

        self._initialized = True

    @property
    def session(self):
        return self.Session()

    def close_session(self):
        self.Session.remove()


class SettingsMixin:
    @property
    def settings_mixin_shared_instance(self):
        return SettingsMixinSharedInstance()

    @property
    def logger(self):
        return self.settings_mixin_shared_instance.logger

    @property
    def session(self):
        return self.settings_mixin_shared_instance.session

    def close_session(self):
        self.settings_mixin_shared_instance.close_session()

    @property
    def stt_settings(self) -> STTSettings:
        return self.load_settings_from_db(STTSettings)

    @property
    def application_settings(self) -> ApplicationSettings:
        return self.load_settings_from_db(ApplicationSettings)

    @property
    def whisper_settings(self) -> WhisperSettings:
        return self.load_settings_from_db(WhisperSettings)

    @property
    def llm_generator_settings(self) -> LLMGeneratorSettings:
        settings = self.load_settings_from_db(LLMGeneratorSettings)
        if settings.current_chatbot == 0:
            chatbots = self.chatbots
            if len(chatbots) > 0:
                settings.current_chatbot = self.chatbots[0].id
                self.update_settings_by_name("llm_generator_settings", "current_chatbot", settings.current_chatbot)
        return settings

    @property
    def generator_settings(self) -> GeneratorSettings:
        return self.load_settings_from_db(GeneratorSettings)

    @property
    def controlnet_settings(self) -> ControlnetSettings:
        return self.load_settings_from_db(ControlnetSettings)

    @property
    def image_to_image_settings(self) -> ImageToImageSettings:
        return self.load_settings_from_db(ImageToImageSettings)

    @property
    def outpaint_settings(self) -> OutpaintSettings:
        return self.load_settings_from_db(OutpaintSettings)

    @property
    def drawing_pad_settings(self) -> DrawingPadSettings:
        return self.load_settings_from_db(DrawingPadSettings)

    @property
    def brush_settings(self) -> BrushSettings:
        return self.load_settings_from_db(BrushSettings)

    @property
    def grid_settings(self) -> GridSettings:
        return self.load_settings_from_db(GridSettings)

    @property
    def active_grid_settings(self) -> ActiveGridSettings:
        return self.load_settings_from_db(ActiveGridSettings)

    @property
    def path_settings(self) -> PathSettings:
        return self.load_settings_from_db(PathSettings)

    @property
    def memory_settings(self) -> MemorySettings:
        return self.load_settings_from_db(MemorySettings)

    @property
    def chatbots(self) -> List[Type[Chatbot]]:
        return self.load_chatbots()

    @property
    def ai_models(self) -> List[Type[AIModels]]:
        return self.load_ai_models()

    @property
    def schedulers(self) -> List[Type[Schedulers]]:
        return self.load_schedulers()

    @property
    def lora(self) -> List[Type[Lora]]:
        return self.load_lora()

    @property
    def shortcut_keys(self) -> List[Type[ShortcutKeys]]:
        return self.load_shortcut_keys()

    @property
    def speech_t5_settings(self) -> SpeechT5Settings:
        return self.load_settings_from_db(SpeechT5Settings)

    @property
    def tts_settings(self) -> TTSSettings:
        return self.load_settings_from_db(TTSSettings)

    @property
    def espeak_settings(self) -> EspeakSettings:
        return self.load_settings_from_db(EspeakSettings)

    @property
    def metadata_settings(self) -> MetadataSettings:
        return self.load_settings_from_db(MetadataSettings)

    @property
    def embeddings(self) -> List[Type[Embedding]]:
        return self.session.query(Embedding).all()

    @property
    def prompt_templates(self) -> List[Type[PromptTemplate]]:
        return self.load_prompt_templates()

    @property
    def controlnet_models(self):
        return self.load_controlnet_models()

    @property
    def saved_prompts(self) -> List[Type[SavedPrompt]]:
        return self.load_saved_prompts()

    @property
    def font_settings(self) -> List[Type[FontSetting]]:
        return self.load_font_settings()

    @property
    def pipelines(self) -> List[Type[PipelineModel]]:
        return self.load_pipelines()

    @property
    def drawing_pad_image(self):
        base_64_image = self.drawing_pad_settings.image
        image = convert_binary_to_image(base_64_image)
        if image is not None:
            image = image.convert("RGB")
        return image

    @property
    def drawing_pad_mask(self):
        base_64_image = self.drawing_pad_settings.mask
        image = convert_binary_to_image(base_64_image)
        if image is not None:
            image = image.convert("RGB")
        return image

    @property
    def img2img_image(self):
        base_64_image = self.image_to_image_settings.image
        image = convert_binary_to_image(base_64_image)
        if image is not None:
            image = image.convert("RGB")
        return image

    @property
    def controlnet_image(self):
        base_64_image = self.controlnet_settings.image
        image = convert_binary_to_image(base_64_image)
        if image is not None:
            image = image.convert("RGB")
        return image

    @property
    def controlnet_generated_image(self):
        base_64_image = self.controlnet_settings.imported_image_base64
        image = convert_binary_to_image(base_64_image)
        if image is not None:
            image = image.convert("RGB")
        return image

    @property
    def outpaint_mask(self):
        base_64_image = self.drawing_pad_settings.mask
        image = convert_binary_to_image(base_64_image)
        if image is not None:
            image = image.convert("RGB")
        return image

    @property
    def image_filter_values(self):
        return self.session.query(ImageFilterValue).all()

    def get_lora_by_version(self, version):
        return self.session.query(Lora).filter_by(version=version).all()

    def get_embeddings_by_version(self, version):
        return [embedding for embedding in self.embeddings if embedding.version == version]

    @property
    def chatbot(self) -> Type[Chatbot]:
        return self.get_chatbot_by_id(
            self.llm_generator_settings.current_chatbot
        )

    @property
    def user(self) -> Type[User]:
        user = self.session.query(User).first()
        if user is None:
            user = User()
            user.username = "User"
            self.session.add(user)
            self.session.commit()
        return user

    @property
    def window_settings(self):
        return self.load_window_settings()

    def add_chatbot_document_to_chatbot(self, chatbot, file_path):
        document = self.session.query(TargetFiles).filter_by(chatbot_id=chatbot.id, file_path=file_path).first()
        if document is None:
            document = TargetFiles(file_path=file_path, chatbot_id=chatbot.id)
        self.session.merge(document)  # Use merge instead of add
        self.session.commit()

    def update_settings_by_name(self, setting_name, column_name, val):
        if setting_name == "application_settings":
            self.update_application_settings(column_name, val)
        elif setting_name == "generator_settings":
            self.update_generator_settings(column_name, val)
        elif setting_name == "controlnet_image_settings":
            self.update_controlnet_image_settings(column_name, val)
        elif setting_name == "brush_settings":
            self.update_brush_settings(column_name, val)
        elif setting_name == "controlnet_settings":
            self.update_controlnet_settings(column_name, val)
        elif setting_name == "image_to_image_settings":
            self.update_image_to_image_settings(column_name, val)
        elif setting_name == "outpaint_settings":
            self.update_outpaint_settings(column_name, val)
        elif setting_name == "drawing_pad_settings":
            self.update_drawing_pad_settings(column_name, val)
        elif setting_name == "grid_settings":
            self.update_grid_settings(column_name, val)
        elif setting_name == "active_grid_settings":
            self.update_active_grid_settings(column_name, val)
        elif setting_name == "path_settings":
            self.update_path_settings(column_name, val)
        elif setting_name == "memory_settings":
            self.update_memory_settings(column_name, val)
        elif setting_name == "llm_generator_settings":
            self.update_llm_generator_settings(column_name, val)
        elif setting_name == "whisper_settings":
            self.update_whisper_settings(column_name, val)
        else:
            logging.error(f"Invalid setting name: {setting_name}")

    def update_application_settings(self, column_name, val):
        self.update_setting(ApplicationSettings, column_name, val)
        self.__settings_updated()

    def update_espeak_settings(self, column_name, val):
        self.update_setting(EspeakSettings, column_name, val)
        self.__settings_updated()

    def update_tts_settings(self, column_name, val):
        self.update_setting(TTSSettings, column_name, val)
        self.__settings_updated()

    def update_speech_t5_settings(self, column_name, val):
        self.update_setting(SpeechT5Settings, column_name, val)
        self.__settings_updated()

    def update_controlnet_settings(self, column_name, val):
        self.update_setting(ControlnetSettings, column_name, val)
        self.__settings_updated()

    def update_brush_settings(self, column_name, val):
        self.update_setting(BrushSettings, column_name, val)
        self.__settings_updated()

    def update_image_to_image_settings(self, column_name, val):
        self.update_setting(ImageToImageSettings, column_name, val)
        self.__settings_updated()

    def update_outpaint_settings(self, column_name, val):
        self.update_setting(OutpaintSettings, column_name, val)
        self.__settings_updated()

    def update_drawing_pad_settings(self, column_name, val):
        self.update_setting(DrawingPadSettings, column_name, val)
        self.__settings_updated()

    def update_grid_settings(self, column_name, val):
        self.update_setting(GridSettings, column_name, val)
        self.__settings_updated()

    def update_active_grid_settings(self, column_name, val):
        self.update_setting(ActiveGridSettings, column_name, val)
        self.__settings_updated()

    def update_path_settings(self, column_name, val):
        self.update_setting(PathSettings, column_name, val)
        self.__settings_updated()

    def update_memory_settings(self, column_name, val):
        self.update_setting(MemorySettings, column_name, val)
        self.__settings_updated()

    def update_metadata_settings(self, column_name, val):
        self.update_setting(MetadataSettings, column_name, val)
        self.__settings_updated()

    def update_llm_generator_settings(self, column_name, val):
        self.update_setting(LLMGeneratorSettings, column_name, val)
        self.__settings_updated()

    def update_whisper_settings(self, column_name, val):
        self.update_setting(WhisperSettings, column_name, val)
        self.__settings_updated()

    def update_ai_models(self, models: List[AIModels]):
        for model in models:
            self.update_ai_model(model)
        self.__settings_updated()

    def update_ai_model(self, model: AIModels):
        query = self.session.query(AIModels).filter_by(
            name=model.name,
            path=model.path,
            branch=model.branch,
            version=model.version,
            category=model.category,
            pipeline_action=model.pipeline_action,
            enabled=model.enabled,
            model_type=model.model_type,
            is_default=model.is_default
        ).first()
        if query:
            for key in model.__dict__.keys():
                if key != "_sa_instance_state":
                    setattr(query, key, getattr(model, key))
        else:
            self.session.add(model)
        self.session.commit()
        self.__settings_updated()

    def update_generator_settings(self, column_name, val):
        generator_settings = self.generator_settings
        setattr(generator_settings, column_name, val)
        self.save_generator_settings(generator_settings)

    def update_controlnet_image_settings(self, column_name, val):
        controlnet_settings = self.controlnet_settings
        setattr(controlnet_settings, column_name, val)
        self.update_controlnet_settings(column_name, val)

    def load_schedulers(self) -> list[Type[Schedulers]]:
        return self.session.query(Schedulers).all()

    def load_settings_from_db(self, model_class_):
        settings = self.session.query(model_class_).first()
        if settings is None:
            settings = self.create_new_settings(model_class_)
        return settings

    def update_setting(self, model_class_, name, value):
        setting = self.session.query(model_class_).order_by(model_class_.id.desc()).first()
        if setting:
            setattr(setting, name, value)
            self.session.commit()

    def save_generator_settings(self, generator_settings: GeneratorSettings):
        query = self.session.query(GeneratorSettings).filter_by(
            id=generator_settings.id
        ).first()
        if query:
            for key in generator_settings.__dict__.keys():
                if key != "_sa_instance_state":
                    setattr(query, key, getattr(generator_settings, key))
        else:
            self.session.add(generator_settings)
        self.session.commit()
        self.__settings_updated()

    def reset_settings(self):
        # Delete all entries from the model class
        self.session.query(ApplicationSettings).delete()
        self.session.query(ActiveGridSettings).delete()
        self.session.query(ControlnetSettings).delete()
        self.session.query(ImageToImageSettings).delete()
        self.session.query(OutpaintSettings).delete()
        self.session.query(DrawingPadSettings).delete()
        self.session.query(MetadataSettings).delete()
        self.session.query(GeneratorSettings).delete()
        self.session.query(LLMGeneratorSettings).delete()
        self.session.query(TTSSettings).delete()
        self.session.query(SpeechT5Settings).delete()
        self.session.query(EspeakSettings).delete()
        self.session.query(STTSettings).delete()
        self.session.query(BrushSettings).delete()
        self.session.query(GridSettings).delete()
        self.session.query(PathSettings).delete()
        self.session.query(MemorySettings).delete()
        # Commit the changes
        self.session.commit()

    def create_new_settings(self, model_class_):
        new_settings = model_class_()
        self.session.add(new_settings)
        self.session.commit()
        self.session.refresh(new_settings)
        return new_settings

    def get_saved_prompt_by_id(self, prompt_id) -> Type[SavedPrompt]:
        return self.session.query(SavedPrompt).filter_by(id=prompt_id).first()

    def update_saved_prompt(self, saved_prompt: SavedPrompt):
        query = self.session.query(SavedPrompt).filter_by(
            id=saved_prompt.id
        ).first()
        if query:
            for key in saved_prompt.__dict__.keys():
                if key != "_sa_instance_state":
                    setattr(query, key, getattr(saved_prompt, key))
        else:
            self.session.add(saved_prompt)
        self.session.commit()
        self.__settings_updated()

    def create_saved_prompt(self, data: dict):
        new_saved_prompt = SavedPrompt(**data)
        self.session.add(new_saved_prompt)
        self.session.commit()

    def load_saved_prompts(self) -> List[Type[SavedPrompt]]:
        return self.session.query(SavedPrompt).all()

    def load_font_settings(self) -> List[Type[FontSetting]]:
        return self.session.query(FontSetting).all()

    def get_font_setting_by_name(self, name) -> Type[FontSetting]:
        return self.session.query(FontSetting).filter_by(name=name).first()

    def update_font_setting(self, font_setting: Type[FontSetting]):
        query = self.session.query(FontSetting).filter_by(
            name=font_setting.name
        ).first()
        if query:
            for key in font_setting.__dict__.keys():
                if key != "_sa_instance_state":
                    setattr(query, key, getattr(font_setting, key))
        else:
            self.session.add(font_setting)
        self.session.commit()
        self.__settings_updated()

    def load_ai_models(self) -> List[Type[AIModels]]:
        return self.session.query(AIModels).all()

    def load_chatbots(self) -> List[Type[Chatbot]]:
        settings = self.session.query(Chatbot).all()
        return settings

    def delete_chatbot_by_name(self, chatbot_name):
        self.session.query(Chatbot).filter_by(name=chatbot_name).delete()
        self.session.commit()

    def create_chatbot(self, chatbot_name):
        new_chatbot = Chatbot(name=chatbot_name)
        self.session.add(new_chatbot)
        self.session.commit()

    def reset_path_settings(self):
        self.session.query(PathSettings).delete()
        self.set_default_values(PathSettings)
        self.session.commit()

    def set_default_values(self, model_name_):
        default_values = {}
        for column in model_name_.__table__.columns:
            if column.default is not None:
                default_values[column.name] = column.default.arg
        self.session.execute(
            model_name_.__table__.insert(),
            [default_values]
        )
        self.session.commit()

    def load_lora(self) -> List[Type[Lora]]:
        return self.session.query(Lora).all()

    def get_lora_by_name(self, name):
        return self.session.query(Lora).filter_by(name=name).first()

    def add_lora(self, lora: Lora):
        self.session.add(lora)
        self.session.commit()

    def delete_lora(self, lora: Lora):
        self.session.query(Lora).filter_by(name=lora.name).delete()
        self.session.commit()

    def update_lora(self, lora: Lora):
        query = self.session.query(Lora).filter_by(name=lora.name).first()
        if query:
            for key in lora.__dict__.keys():
                if key != "_sa_instance_state":
                    setattr(query, key, getattr(lora, key))
        else:
            self.session.add(lora)
        self.session.commit()
        self.__settings_updated()

    def update_loras(self, loras: List[Lora]):
        for lora in loras:
            query = self.session.query(Lora).filter_by(name=lora.name).first()
            if query:
                for key in lora.__dict__.keys():
                    if key != "_sa_instance_state":
                        setattr(query, key, getattr(lora, key))
            else:
                self.session.add(lora)
        self.session.commit()
        self.__settings_updated()

    def create_lora(self, lora: Lora):
        self.session.add(lora)
        self.session.commit()

    def delete_lora_by_name(self, lora_name, version):
        self.session.query(Lora).filter_by(name=lora_name, version=version).delete()
        self.session.commit()

    def delete_embedding(self, embedding: Embedding):
        self.session.query(Embedding).filter_by(
            name=embedding.name,
            path=embedding.path,
            branch=embedding.branch,
            version=embedding.version,
            category=embedding.category,
            pipeline_action=embedding.pipeline_action,
            enabled=embedding.enabled,
            model_type=embedding.model_type,
            is_default=embedding.is_default
        ).delete()
        self.session.commit()

    def update_embeddings(self, embeddings: List[Embedding]):
        for embedding in embeddings:
            query = self.session.query(Embedding).filter_by(
                name=embedding.name,
                path=embedding.path,
                branch=embedding.branch,
                version=embedding.version,
                category=embedding.category,
                pipeline_action=embedding.pipeline_action,
                enabled=embedding.enabled,
                model_type=embedding.model_type,
                is_default=embedding.is_default
            ).first()
            if query:
                for key in embedding.__dict__.keys():
                    if key != "_sa_instance_state":
                        setattr(query, key, getattr(embedding, key))
            else:
                self.session.add(embedding)
        self.session.commit()
        self.__settings_updated()

    def get_embedding_by_name(self, name):
        return self.session.query(Embedding).filter_by(name=name).first()

    def add_embedding(self, embedding: Embedding):
        self.session.add(embedding)
        self.session.commit()

    def load_prompt_templates(self) -> List[Type[PromptTemplate]]:
        return self.session.query(PromptTemplate).all()

    def get_prompt_template_by_name(self, name) -> Type[PromptTemplate]:
        return self.session.query(PromptTemplate).filter_by(template_name=name).first()

    def load_controlnet_models(self) -> List[Type[ControlnetModel]]:
        return self.session.query(ControlnetModel).all()

    def controlnet_model_by_name(self, name) -> Type[ControlnetModel]:
        return self.session.query(ControlnetModel).filter_by(name=name).first()

    def load_pipelines(self) -> List[Type[PipelineModel]]:
        return self.session.query(PipelineModel).all()

    def load_shortcut_keys(self) -> List[Type[ShortcutKeys]]:
        return self.session.query(ShortcutKeys).all()

    def load_window_settings(self) -> Type[WindowSettings]:
        return self.session.query(WindowSettings).first()

    def save_window_settings(self, column_name, val):
        window_settings = self.window_settings
        setattr(window_settings, column_name, val)
        query = self.session.query(WindowSettings).first()
        if query:
            for key in window_settings.__dict__.keys():
                if key != "_sa_instance_state":
                    setattr(query, key, getattr(window_settings, key))
        else:
            self.session.add(window_settings)
        self.session.commit()

    def save_object(self, database_object):
        self.session.add(database_object)
        self.session.commit()

    def load_history_from_db(self, conversation_id):
        messages = self.session.query(Message).filter_by(
            conversation_id=conversation_id
        ).order_by(Message.timestamp).all()
        results = [
            {
                "id": message.id,
                "role": message.role,
                "content": message.content,
                "name": message.name,
                "is_bot": message.is_bot,
                "timestamp": message.timestamp,
                "conversation_id": message.conversation_id
            } for message in messages
        ]
        return results

    def save_message(self, content, role, name, is_bot, conversation_id) -> Message:
        timestamp = datetime.datetime.now()  # Ensure timestamp is a datetime object
        llm_generator_settings = self.session.query(LLMGeneratorSettings).first()
        message = Message(
            role=role,
            content=content,
            name=name,
            is_bot=is_bot,
            timestamp=timestamp,
            conversation_id=conversation_id,
            chatbot_id=llm_generator_settings.current_chatbot
        )
        self.session.add(message)
        self.session.commit()
        return message

    def get_chatbot_by_id(self, chatbot_id) -> Type[Chatbot]:
        chatbot = self.session.query(Chatbot).filter_by(id=chatbot_id).options(joinedload(Chatbot.target_files)).first()
        if chatbot is None:
            chatbot = self.session.query(Chatbot).options(joinedload(Chatbot.target_files)).first()
        return chatbot

    def create_conversation(self):
        # find conversation which has no title, bot_mood or messages
        conversation = self.session.query(Conversation).filter_by(title="", bot_mood="").first()
        if conversation:
            # ensure there are no messages in the conversation
            message = self.session.query(Message).filter_by(conversation_id=conversation.id).first()
            if message is None:
                return conversation
        conversation = Conversation(
            timestamp=datetime.datetime.now(datetime.timezone.utc),
            title=""
        )
        self.session.add(conversation)
        self.session.commit()
        return conversation

    def update_conversation_title(self, conversation_id, title):
        conversation = self.session.query(Conversation).filter_by(id=conversation_id).first()
        if conversation:
            conversation.title = title
            self.session.commit()

    def add_summary(self, content, conversation_id):
        timestamp = datetime.datetime.now()  # Ensure timestamp is a datetime object
        summary = Summary(
            content=content,
            timestamp=timestamp,
            conversation_id=conversation_id
        )
        self.session.add(summary)
        self.session.commit()

    def get_all_conversations(self):
        conversations = self.session.query(Conversation).all()
        return conversations

    def delete_conversation(self, conversation_id):
        self.session.query(Message).filter_by(conversation_id=conversation_id).delete()
        self.session.query(Summary).filter_by(conversation_id=conversation_id).delete()
        self.session.query(Conversation).filter_by(id=conversation_id).delete()
        self.session.commit()

    def get_most_recent_conversation(self):
        conversation = self.session.query(Conversation).order_by(Conversation.timestamp.desc()).first()
        return conversation

    def __settings_updated(self):
        self.emit_signal(SignalCode.APPLICATION_SETTINGS_CHANGED_SIGNAL)



from PySide6.QtCore import Slot, QTimer

from airunner.enums import SignalCode, StableDiffusionVersion, ImageGenerator, GeneratorSection
from airunner.utils.create_worker import create_worker
from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.model_manager.model_widget import ModelWidget
from airunner.widgets.model_manager.templates.custom_ui import Ui_custom_model_widget

from PySide6 import QtWidgets

from airunner.windows.main.ai_model_mixin import AIModelMixin
from airunner.windows.main.pipeline_mixin import PipelineMixin
from airunner.workers.model_scanner_worker import ModelScannerWorker


class CustomModelWidget(
    BaseWidget,
    PipelineMixin,
    AIModelMixin
):
    widget_class_ = Ui_custom_model_widget
    model_widgets = []
    spacer = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        PipelineMixin.__init__(self)
        AIModelMixin.__init__(self)
        self.initialized = False
        self.model_scanner_worker = create_worker(ModelScannerWorker)
        self.register(SignalCode.DOWNLOAD_COMPLETE, self.scan_for_models)

    def showEvent(self, event):
        super().showEvent(event)
        QTimer.singleShot(
            50,
            self.__do_show
        )

    def __do_show(self):
        if not self.initialized:
            self.show_items_in_scrollarea()
            self.initialized = True

    @Slot()
    def action_button_clicked_scan_for_models(self):
        self.scan_for_models()

    def scan_for_models(self, data=None):
        self.model_scanner_worker.add_to_queue("scan_for_models")

    def show_items_in_scrollarea(self, search=None):
        if self.spacer:
            self.ui.scrollAreaWidgetContents.layout().removeItem(self.spacer)
        for child in self.ui.scrollAreaWidgetContents.children():
            if isinstance(child, ModelWidget):
                child.deleteLater()
        if search:
            models = self.ai_models_find(search, default=False)
        else:
            models = self.ai_models_find(default=False)
        for model_widget in self.model_widgets:
            model_widget.deleteLater()
        self.model_widgets = []
        for index, model in enumerate(models):
            version = model.get("version", StableDiffusionVersion.SD1_5)
            category = model.get("category", ImageGenerator.STABLEDIFFUSION)
            pipeline_action = model.get("pipeline_action", GeneratorSection.TXT2IMG)
            pipeline_class = self.get_pipeline_classname(
                pipeline_action, version, category)
            model_widget = ModelWidget(
                path=model["path"],
                branch=model["branch"],
                version=version,
                category=category,
                pipeline_action=pipeline_action,
                pipeline_class=pipeline_class,
            )

            model_widget.ui.name.setChecked(model["enabled"])

            self.ui.scrollAreaWidgetContents.layout().addWidget(
                model_widget)

            self.model_widgets.append(model_widget)
        
        if not self.spacer:
            self.spacer = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.ui.scrollAreaWidgetContents.layout().addItem(self.spacer)

    def mode_type_changed(self, val):
        print("mode_type_changed", val)
    
    def toggle_all_toggled(self, val):
        print("toggle_all_toggled", val)
    
    def search_text_edited(self, val):
        val = val.strip()
        if val == "":
            val = None
        self.show_items_in_scrollarea(val)



from airunner.windows.setup_wizard.user_agreement.agreement_page import AgreementPage
from airunner.windows.setup_wizard.user_agreement.templates.user_agreement_ui import Ui_user_agreement


class UserAgreement(AgreementPage):
    class_name_ = Ui_user_agreement
    setting_key = "user_agreement_checked"




from PySide6.QtCore import Slot

from airunner.data.models.settings_models import PromptTemplate
from airunner.settings import DEFAULT_IMAGE_SYSTEM_PROMPT, DEFAULT_APPLICATION_COMMAND_SYSTEM_PROMPT, \
    DEFAULT_UPDATE_MOOD_SYSTEM_PROMPT, DEFAULT_RAG_SEARCH_SYSTEM_PROMPT, DEFAULT_CHATBOT_SYSTEM_PROMPT, \
    DEFAULT_SUMMARIZE_CHAT_SYSTEM_PROMPT, DEFAULT_IMAGE_LLM_GUARDRAILS, DEFAULT_CHATBOT_GUARDRAILS_PROMPT
from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.llm.templates.prompt_templates_ui import Ui_prompt_templates_widget


class PromptTemplatesWidget(BaseWidget):
    widget_class_ = Ui_prompt_templates_widget

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        session = self.session
        self._prompt_templates = session.query(PromptTemplate).all()
        self.ui.template_name.clear()
        self.current_template_index = 0
        for template in self._prompt_templates:
            self.ui.template_name.addItem(template.template_name)
        self.ui.template_name.setCurrentIndex(0)

    @Slot(int)
    def template_changed(self, index:int):
        self.current_template_index = index
        template = self._prompt_templates[index]
        self.ui.system_prompt.blockSignals(True)
        self.ui.guardrails_prompt.blockSignals(True)
        self.ui.use_guardrails.blockSignals(True)
        self.ui.use_datetime.blockSignals(True)
        self.ui.system_prompt.setPlainText(template.system)
        self.ui.guardrails_prompt.setPlainText(template.guardrails)
        self.ui.use_guardrails.setChecked(template.use_guardrails)
        self.ui.use_datetime.setChecked(template.use_system_datetime_in_system_prompt)
        self.ui.system_prompt.blockSignals(False)
        self.ui.guardrails_prompt.blockSignals(False)
        self.ui.use_guardrails.blockSignals(False)
        self.ui.use_datetime.blockSignals(False)

    @Slot()
    def system_prompt_changed(self):
        session = self.session
        template = self._prompt_templates[self.current_template_index]
        template.system = self.ui.system_prompt.toPlainText()
        self._prompt_templates[self.current_template_index] = template
        session.commit()

    @Slot()
    def guardrails_prompt_changed(self):
        session = self.session
        template = self._prompt_templates[self.current_template_index]
        template.guardrails = self.ui.guardrails_prompt.toPlainText()
        self._prompt_templates[self.current_template_index] = template
        session.commit()

    @Slot(bool)
    def toggle_use_guardrails(self, val:bool):
        session = self.session
        template = self._prompt_templates[self.current_template_index]
        template.use_guardrails = val
        self._prompt_templates[self.current_template_index] = template
        session.commit()

    @Slot(bool)
    def toggle_use_datetime(self, val:bool):
        session = self.session
        template = self._prompt_templates[self.current_template_index]
        template.use_system_datetime_in_system_prompt = val
        self._prompt_templates[self.current_template_index] = template
        session.commit()

    Slot()
    def reset_system_prompt(self):
        session = self.session
        template = self._prompt_templates[self.current_template_index]

        if template.template_name == "image":
            default = DEFAULT_IMAGE_SYSTEM_PROMPT
        elif template.template_name == "application_command":
            default = DEFAULT_APPLICATION_COMMAND_SYSTEM_PROMPT
        elif template.template_name == "update_mood":
            default = DEFAULT_UPDATE_MOOD_SYSTEM_PROMPT
        elif template.template_name == "rag_search":
            default = DEFAULT_RAG_SEARCH_SYSTEM_PROMPT
        elif template.template_name == "chatbot":
            default = DEFAULT_CHATBOT_SYSTEM_PROMPT
        elif template.template_name == "summarize":
            default = DEFAULT_SUMMARIZE_CHAT_SYSTEM_PROMPT
        else:
            default = ""

        template.system = default
        self.ui.system_prompt.setPlainText(default)
        self._prompt_templates[self.current_template_index] = template
        session.commit()

    Slot()
    def reset_guardrails_prompt(self):
        session = self.session
        template = self._prompt_templates[self.current_template_index]

        if template.template_name == "image":
            default = DEFAULT_IMAGE_LLM_GUARDRAILS
        elif template.template_name == "chatbot":
            default = DEFAULT_CHATBOT_GUARDRAILS_PROMPT
        else:
            default = ""

        template.guardrails = default
        self.ui.guardrails_prompt.setPlainText(default)
        self._prompt_templates[self.current_template_index] = template
        session.commit()



import random
import os

import torch

from llama_index.llms.groq import Groq

from transformers.utils.quantization_config import BitsAndBytesConfig, GPTQConfig
from transformers import AutoModelForCausalLM, AutoTokenizer
from transformers.generation.streamers import TextIteratorStreamer

from airunner.handlers.base_handler import BaseHandler
from airunner.enums import SignalCode, ModelType, ModelStatus, LLMActionType
from airunner.settings import MAX_SEED
from airunner.utils.clear_memory import clear_memory
from airunner.handlers.llm.agent.base_agent import BaseAgent


class CausalLMTransformerBaseHandler(
    BaseHandler
):
    auto_class_ = AutoModelForCausalLM
    tokenizer_class_ = AutoTokenizer
    model_type = ModelType.LLM

    def __init__(self, *args, **kwargs):
        self.model_type = ModelType.LLM
        self.model_class = "llm"
        self.agent_options = kwargs.pop("agent_options", {})
        self._model = None
        self._streamer = None
        self._chat_engine = None
        self._chat_agent = None
        self._llm_with_tools = None
        self._agent_executor = None
        self._embed_model = None
        self._service_context_model = None
        self._use_query_engine: bool = False
        self._use_chat_engine: bool = True
        self._user_evaluation: str = ""
        self._restrict_tools_to_additional: bool = True
        self._return_agent_code: bool = False
        self._rag_tokenizer = None
        self._rag_retriever = None
        self._do_quantize_model = kwargs.pop("do_quantize_model", True)
        self.__model = None
        self._vocoder = None
        self._current_model_path = kwargs.get("current_model_path", "")
        self._history = []
        self._set_attention_mask = kwargs.get("set_attention_mask", False)
        self._do_push_to_hub = kwargs.get("do_push_to_hub", False)
        self._llm_int8_enable_fp32_cpu_offload = kwargs.get("llm_int8_enable_fp32_cpu_offload", True)
        self._generator_name = kwargs.get("generator_name", "")
        self._return_result = kwargs.get("return_result", True)
        self._skip_special_tokens = kwargs.get("skip_special_tokens", True)
        self._processing_request = kwargs.get("_processing_request", False)
        self._bad_words_ids = kwargs.get("bad_words_ids", None)
        self._bos_token_id = kwargs.get("bos_token_id", None)
        self._pad_token_id = kwargs.get("pad_token_id", None)
        self._eos_token_id = kwargs.get("eos_token_id", None)
        self._no_repeat_ngram_size = kwargs.get("no_repeat_ngram_size", 1)
        self._decoder_start_token_id = kwargs.get("decoder_start_token_id", None)
        self._tokenizer = None
        self._generator = None

        super().__init__(*args, **kwargs)

    @property
    def is_mistral(self) -> bool:
        path = self._current_model_path.lower()
        return "mistral" in path

    @property
    def is_llama_instruct(self):
        path = self._current_model_path.lower()
        if "instruct" in path and "llama" in path:
            return True
        return False

    @property
    def chat_template(self):
        if self.is_mistral:
            return (
                "{% for message in messages %}"
                "{% if message['role'] == 'system' %}"
                "{{ '[INST] <<SYS>>' + message['content'] + ' <</SYS>>[/INST]' }}"
                "{% elif message['role'] == 'user' %}"
                "{{ '[INST]' + message['content'] + ' [/INST]' }}"
                "{% elif message['role'] == 'assistant' %}"
                "{{ message['content'] + eos_token + ' ' }}"
                "{% endif %}"
                "{% endfor %}"
            )
        elif self.is_llama_instruct:
            return (
                "{{ '<|begin_of_text|>' }}"
                "{% for message in messages %}"
                "{{ '<|start_header_id|>' + "
                "message['role'] + '<|end_header_id|>' + '\n\n' + message['content'] + "
                "'<|end_header_id|>\n\n' + message['content'] + '<|eot_id|>' }}"
                "{% endfor %}"
            )

    @property
    def username(self):
        return self.user.username

    @property
    def botname(self):
        if self.chatbot.assign_names:
            return self.chatbot.botname
        return "Assistant"

    @property
    def _quantization_config(self):
        config = None
        if self.llm_dtype == "8bit":
            config = BitsAndBytesConfig(
                load_in_4bit=False,
                load_in_8bit=True,
                llm_int8_threshold=6.0,
                llm_int8_has_fp16_weight=False,
                bnb_4bit_compute_dtype=torch.bfloat16,
                bnb_4bit_use_double_quant=True,
                bnb_4bit_quant_type='nf4',
            )
        elif self.llm_dtype == "4bit":
            config = BitsAndBytesConfig(load_in_4bit=True)
        elif self.llm_dtype == "2bit":
            config = GPTQConfig(
                bits=2,
                dataset="c4",
                tokenizer=self._tokenizer
            )
        return config

    @property
    def use_cache(self):
        if self.llm_generator_settings.override_parameters:
            return self.llm_generator_settings.use_cache
        return self.chatbot.use_cache

    @property
    def model_path(self):
        model_version:str = self.chatbot.model_version
        if self.llm_generator_settings.override_parameters:
            model_version = self.llm_generator_settings.model_version
        current_llm_generator = self.application_settings.current_llm_generator
        local_path:str = "misc"
        if current_llm_generator == "causallm":
            local_path = "causallm"
        elif current_llm_generator == "seq2seq":
            local_path = "seq2seq"
        elif current_llm_generator == "visualqa":
            local_path = "visualqa"
        base:str = self.path_settings.base_path
        return os.path.expanduser(os.path.join(
            base,
            "text",
            "models",
            "llm",
            local_path,
            model_version
        ))

    def load(self):
        if self.model_status in (
            ModelStatus.LOADING,
            ModelStatus.LOADED
        ):
            return
        self.unload()
        self.change_model_status(ModelType.LLM, ModelStatus.LOADING)
        self._current_model_path = self.model_path
        self._load_tokenizer()
        self._load_model()
        self._load_streamer()
        self._load_agent()
        if self._model and self._tokenizer and self._streamer and self._chat_agent:
            self.change_model_status(ModelType.LLM, ModelStatus.LOADED)
        else:
            self.change_model_status(ModelType.LLM, ModelStatus.FAILED)

    def unload(self):
        if self.model_status in (
            ModelStatus.LOADING,
            ModelStatus.UNLOADED
        ):
            return
        self.logger.debug("Unloading LLM")
        self.change_model_status(ModelType.LLM, ModelStatus.LOADING)
        self._unload_streamer()
        self._unload_llm_with_tools()
        self._unload_agent_executor()
        self._unload_embed_model()
        self._unload_model()
        self._unload_tokenizer()
        self._unload_agent()
        clear_memory()
        self.change_model_status(ModelType.LLM, ModelStatus.UNLOADED)

    def handle_request(self, data:dict):
        self.logger.debug("Handling request")
        self._processing_request = True
        self._do_set_seed()
        self.load()
        self._processing_request = True
        action = self.llm_generator_settings.action
        if type(action) is str:
            action = LLMActionType[action]
        self._do_generate(
            data["request_data"]["prompt"],
            action
        )

    def do_interrupt(self):
        """
        Public method to interrupt the chat process
        """
        if self._chat_agent:
            self._chat_agent.interrupt_process()

    def clear_history(self):
        """
        Public method to clear the chat agent history
        """
        if not self._chat_agent:
            return
        self.logger.debug("Clearing chat history")
        self._chat_agent.clear_history()

    def add_chatbot_response_to_history(self, message):
        """
        Public method to add a chatbot response to the chat agent history
        """
        self._chat_agent.add_chatbot_response_to_history(message)

    def load_conversation(self, message):
        """
        Public method to load a conversation into the chat agent
        """
        self._chat_agent.on_load_conversation(message)

    def reload_rag(self):
        """
        Public method to reload the RAG model
        """
        self._chat_agent.reload_rag()

    def _load_tokenizer(self):
        if self._tokenizer is not None:
            return
        path = self.model_path
        self.logger.debug(f"Loading tokenizer from {path}")
        kwargs = {
            "local_files_only": True,
            "device_map": self.device,
            "trust_remote_code": True,
            "torch_dtype": self.torch_dtype,
            "attn_implementation": "flash_attention_2",
        }

        if self.chat_template:
            kwargs["chat_template"] = self.chat_template
        try:
            self._tokenizer = self.tokenizer_class_.from_pretrained(
                path,
                **kwargs,
            )
            self.logger.debug("Tokenizer loaded")
        except Exception as e:
            self.logger.error(e)

        if self._tokenizer:
            self._tokenizer.use_default_system_prompt = False
        else:
            self.logger.error("Tokenizer failed to load")

    def _load_model(self):
        if self._model is not None:
            return
        self.logger.debug("transformer_base_handler.load_model Loading model")
        if self.llm_generator_settings.use_api:
            self._model = Groq(
                model=self.llm_generator_settings.api_model,
                api_key=self.llm_generator_settings.api_key,
            )
        else:
            self._load_model_local()

    def _load_streamer(self):
        if self._streamer is not None:
            return
        self.logger.debug("Loading LLM text streamer")
        self._streamer = TextIteratorStreamer(self._tokenizer)

    def _load_agent(self):
        if self._chat_agent is not None:
            return
        self.logger.debug("Loading agent")
        self._chat_agent = BaseAgent(
            model=self._model,
            tokenizer=self._tokenizer,
            chat_template=self.chat_template,
            is_mistral=self.is_mistral,
        )

    def _unload_streamer(self):
        self.logger.debug("Unloading streamer")
        del self._streamer
        self._streamer = None

    def _unload_llm_with_tools(self):
        self.logger.debug("Unloading LLM with tools")
        del self._llm_with_tools
        self._llm_with_tools = None

    def _unload_agent_executor(self):
        self.logger.debug("Unloading agent executor")
        del self._agent_executor
        self._agent_executor = None

    def _unload_embed_model(self):
        self.logger.debug("Unloading embed model")
        del self._embed_model
        self._embed_model = None

    def _unload_model(self):
        self.logger.debug("Unloading model")
        self._model = None
        return True

    def _unload_tokenizer(self):
        self.logger.debug("Unloading tokenizer")
        del self._tokenizer
        self._tokenizer = None
        clear_memory(self.memory_settings.default_gpu_llm)
        return True

    def _unload_agent(self):
        self.logger.debug("Unloading agent")
        do_clear_memory = False
        if self._chat_agent is not None:
            self.logger.debug("Unloading chat agent")
            self._chat_agent.unload()
            del self._chat_agent
            self._chat_agent = None
            do_clear_memory = True
        return do_clear_memory

    def _load_model_local(self):
        self.logger.debug("Loading local LLM model")
        path = self.model_path
        is_quantized = os.path.exists(path)
        if not is_quantized:
            path = self.model_path
        params = dict(
            local_files_only=True,
            use_cache=self.use_cache,
            trust_remote_code=False,
            torch_dtype=self.torch_dtype,
            device_map=self.device,
        )
        if self._do_quantize_model and self.use_cuda:
            config = self._quantization_config
            if config:
                params["quantization_config"] = config
        try:
            self._model = self.auto_class_.from_pretrained(
                path,
                **params
            )
        except Exception as e:
            self.logger.error(f"Error loading model: {e}")
            self._model = None

    def _do_generate(self, prompt: str, action: LLMActionType):
        self.logger.debug("Generating response")
        if self._current_model_path != self.model_path:
            self.unload()
            self.load()
        if action is LLMActionType.CHAT and self.chatbot.use_mood:
            action = LLMActionType.UPDATE_MOOD
        self._chat_agent.run(
            prompt,
            action
        )
        if action is LLMActionType.CHAT:
            self._send_final_message()

    def _emit_streamed_text_signal(self, **kwargs):
        self.logger.debug("Emitting streamed text signal")
        kwargs["name"] = self.botname
        self.emit_signal(
            SignalCode.LLM_TEXT_STREAMED_SIGNAL,
            kwargs
        )

    def _send_final_message(self):
        self.logger.debug("Sending final message")
        self._emit_streamed_text_signal(
            message="",
            is_first_message=False,
            is_end_of_message=True
        )

    def _do_set_seed(self):
        self.logger.debug("Setting seed")

        if self.llm_generator_settings.override_parameters:
            seed = self.llm_generator_settings.seed
            random_seed = self.llm_generator_settings.random_seed
        else:
            seed = self.chatbot.seed
            random_seed = self.chatbot.random_seed

        if random_seed:
            seed = random.randint(-MAX_SEED, MAX_SEED)

        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        random.seed(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        if self._tokenizer:
            self._tokenizer.seed = seed

    def _save_quantized_model(self):
        self.logger.debug("Saving quantized model to cache")
        self._model.save_pretrained(self.model_path)

    def _clear_memory(self):
        self.logger.debug("Clearing memory")
        clear_memory(self.memory_settings.default_gpu_llm)



from airunner.enums import SignalCode
from airunner.widgets.base_widget import BaseWidget
from airunner.windows.prompt_browser.templates.prompt_browser_prompt_widget_ui import Ui_prompt_widget


class PromptWidget(BaseWidget):
    widget_class_ = Ui_prompt_widget

    def __init__(self, *args, **kwargs):
        self.saved_prompt = kwargs.pop("saved_prompt")
        super().__init__(*args, **kwargs)
        self.ui.prompt.blockSignals(True)
        self.ui.negative_prompt.blockSignals(True)
        self.ui.secondary_prompt.blockSignals(True)
        self.ui.secondary_negative_prompt.blockSignals(True)
        self.ui.prompt.setPlainText(self.saved_prompt.prompt)
        self.ui.negative_prompt.setPlainText(self.saved_prompt.negative_prompt)
        self.ui.secondary_prompt.setPlainText(self.saved_prompt.secondary_prompt)
        self.ui.secondary_negative_prompt.setPlainText(self.saved_prompt.secondary_negative_prompt)
        self.ui.prompt.blockSignals(False)
        self.ui.negative_prompt.blockSignals(False)
        self.ui.secondary_prompt.blockSignals(False)
        self.ui.secondary_negative_prompt.blockSignals(False)

    def action_text_changed_prompt(self):
        self.save_prompt()

    def action_text_changed_negative_prompt(self):
        self.save_prompt()

    def action_clicked_button_load(self):
        self.emit_signal(SignalCode.SD_LOAD_PROMPT_SIGNAL, {
            "saved_prompt": self.saved_prompt
        })

    def action_clicked_button_delete(self):
        
        self.session.delete(self.saved_prompt)
        self.session.commit()
        
        self.deleteLater()

    def save_prompt(self):
        self.saved_prompt.prompt = self.ui.prompt.toPlainText()
        self.saved_prompt.negative_prompt = self.ui.negative_prompt.toPlainText()
        self.saved_prompt.secondary_prompt = self.ui.secondary_prompt.toPlainText()
        self.saved_prompt.secondary_negative_prompt = self.ui.secondary_negative_prompt.toPlainText()
        self.update_saved_prompt(self.saved_prompt)



from PySide6.QtCore import Qt, QObject
from PySide6.QtWidgets import QColorDialog


class ColorPicker(QColorDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowFlags(Qt.WindowType.Widget)
        self.setOptions(
            QColorDialog.ColorDialogOption.DontUseNativeDialog |
            QColorDialog.ColorDialogOption.NoButtons
        )
        self.delete_elements(self)
        try:
            color_picker_widget = self.findChild(QObject, "color_picker")
            # change this layout to qgridlayout
            # remove color_picker_widget from its parent
            color_picker_widget.setParent(None)
            # delete all children from this widget
            self.delete_elements(self)
            # add grid_layout to this widget
            while self.layout().count():
                item = self.layout().takeAt(0)
                if item.widget():
                    item.widget().deleteLater()
            self.layout().addWidget(color_picker_widget)
        except Exception as e:
            print(e)

    def delete_elements(self, widget):
        # iterate over all children in layout and print their label names
        try:
            for child in widget.children():
                class_name = child.metaObject().className()
                if class_name in [
                    "QGridLayout", "QVBoxLayout",
                    "QWellArray", "QPushButton", "QDialogButtonBox", "QLabel",
                ]:
                    child.deleteLater()
                elif class_name == "QColorPicker":
                    # set a label name
                    child.setObjectName("color_picker")
                self.delete_elements(child)
        except Exception as e:
            print(e)



from typing import List

from airunner.enums import SignalCode


class AIModelMixin:
    def on_ai_model_delete_signal(self, item: dict):
        self.ai_models = [existing_item for existing_item in self.ai_models if existing_item.name != item.name]
        self.update_settings("ai_models", self.ai_models)

    def ai_model_names_by_section(self, section) -> List[str]:
        return [model.name for model in self.ai_models if model.section == section]

    def models_by_pipeline_action(self, pipeline_action: str):
        val = [model for model in self.ai_models if model.pipeline_action == pipeline_action]
        return val
    
    def ai_models_find(self, search="", default=False):
        return [model for model in self.ai_models if model.is_default == default and search.lower() in model.name.lower()]

    def ai_model_get_disabled_default(self):
        return [model for model in self.ai_models if model.is_default is True and model.enabled is False]

    def on_ai_models_save_or_update_signal(self, data: dict):
        new_models = data.get("models", [])
        default_models = self.ai_models

        # Convert list of models to dictionary with model name as key
        model_dict = {model.name: model for model in default_models}

        # Update the dictionary with existing models
        # model_dict.update({model['name']: model for model in existing_models})

        # Update the dictionary with new models
        model_dict.update({model.name: model for model in new_models})

        # Convert back to list
        merged_models = list(model_dict.values())

        self.update_ai_models(merged_models)
        self.emit_signal(SignalCode.AI_MODELS_CREATE_SIGNAL)
        
    def ai_model_paths(self, model_type=None, pipeline_action=None):
        models = self.ai_models
        if model_type:
            models = [model for model in models if model.model_type == model_type]
        if pipeline_action:
            models = [model for model in models if model.pipeline_action == pipeline_action]

        return [model.path for model in models]

    def ai_model_categories(self):
        return [model.category for model in self.ai_models]
    
    def ai_model_pipeline_actions(self):
        return [model.pipeline_action for model in self.ai_models]
    
    def ai_model_versions(self):
        return [model.version for model in self.ai_models]
    
    def ai_models_by_category(self, category):
        return [model for model in self.ai_models if model.category == category]

    def ai_model_get_all(self):
        return self.ai_models



from PySide6 import QtWidgets

from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.model_manager.templates.model_form_ui import Ui_model_form_widget

class ModelFormWidget(BaseWidget):
    widget_class_ = Ui_model_form_widget
    model_widgets = []

    def set_model_form_data(
        self, 
        categories, 
        actions, 
        diffuser_model_versions, 
        category, 
        pipeline_action, 
        pipeline_class, 
        diffuser_model_version, 
        path, 
        model_name,
        model_data,
        model_type
    ):
        self.ui.category.clear()
        self.ui.category.addItems(categories)
        self.ui.category.setCurrentText(category)
        self.ui.pipeline_action.clear()
        self.ui.pipeline_action.addItems(actions)
        self.ui.pipeline_action.setCurrentText(pipeline_action)
        self.ui.model_name.setText(model_name)
        self.ui.diffuser_model_version.clear()
        self.ui.diffuser_model_version.addItems(diffuser_model_versions)
        self.ui.diffuser_model_version.setCurrentText(diffuser_model_version)
        self.ui.model_type.clear()
        self.ui.model_type.addItems(["Checkpoint", "LORA", "Embedding", "VAE", "Controlnet", "Pose"])
        self.ui.enabled.setChecked(True)
        self.ui.path_line_edit.setText(path)

        # set current model type
        if model_type == "TextualInversion":
            model_type = "Embedding"
        self.ui.model_type.setCurrentText(model_type)

        allowCommercialUse = model_data["allowCommercialUse"]
        if type(allowCommercialUse) == list:
            allowCommercialUse = ", ".join(allowCommercialUse)

        # self.ui.model_data.show()
        #
        # if data is None:
        #     self.ui.model_data.hide()

        self.ui.poi.setText(str(model_data["poi"]))
        self.ui.require_credit.setText(str(model_data["allowNoCredit"]))
        self.ui.allow_commercial_use.setText(str(allowCommercialUse))
        self.ui.allow_derivatives.setText(str(model_data["allowDerivatives"]))
        self.ui.allow_different_license.setText(str(model_data["allowDifferentLicense"]))


#!/usr/bin/env python
"""
----------------------------------------------------------------
Import order is crucial for AI Runner to work as expected.
Do not remove the no_internet_socket import.
Do not change the order of the imports.
----------------------------------------------------------------
"""
################################################################
# Importing this module sets the Hugging Face environment
# variables for the application.
################################################################
# import facehuggershield
from airunner.settings import NLTK_DOWNLOAD_DIR
import os
base_path = os.path.join(os.path.expanduser("~"), ".local", "share", "airunner")
# facehuggershield.huggingface.activate(
#     show_stdout=True,
#     darklock_os_whitelisted_directories=[
#         base_path,
#         NLTK_DOWNLOAD_DIR,
#         "/tmp"
#     ]
# )

################################################################
# Set the environment variable for PyTorch to use expandable
################################################################
os.environ["PYTORCH_CUDA_ALLOC_CONF"] = "expandable_segments:True"

################################################################
# Ensure that the base directory exists.
################################################################
base_dir = os.path.join(base_path, "data")
os.makedirs(base_dir, exist_ok=True)

################################################################
# Import the main application class for AI Runner.
################################################################
from airunner.app import App

###############################################################
# Import Alembic modules to run migrations.
################################################################
from alembic.config import Config
from alembic import command
from pathlib import Path
from airunner.data.models.settings_models import ApplicationSettings
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session

def setup_database():
    base_path = Path(os.path.dirname(os.path.realpath(__file__)))
    alembic_file = base_path / "alembic.ini"
    alembic_dir = base_path / "alembic"
    alembic_cfg = Config(alembic_file)
    alembic_cfg.set_main_option("script_location", str(alembic_dir))
    command.upgrade(alembic_cfg, "head")

def main():
    setup_database()

    # Get the first ApplicationSettings record from the database and check for run_setup_wizard boolean
    engine = create_engine("sqlite:///" + os.path.join(base_dir, "airunner.db"))
    session = scoped_session(sessionmaker(bind=engine))
    application_settings = session.query(ApplicationSettings).first()
    if application_settings.run_setup_wizard:
        from airunner.app_installer import AppInstaller
        AppInstaller()
    else:
        App()


if __name__ == "__main__":
    main()



from typing import List, Dict, AnyStr, Any
from PIL import Image, PngImagePlugin
import os

def export_image(
    image: Image,
    file_path: AnyStr,
    metadata: Dict = None
):
    base, ext = os.path.splitext(file_path)
    current_path = file_path
    counter = 1
    while os.path.exists(current_path):
        current_path = f"{base}_{counter}{ext}"
        counter += 1

    if metadata and ext.lower() == '.png':
        png_info = PngImagePlugin.PngInfo()
        for key, value in metadata.items():
            png_info.add_text(key, str(value))
        image.save(current_path, pnginfo=png_info)
    else:
        image.save(current_path)

def export_images(
    images: List[Any],
    file_path: AnyStr,
    metadata: List[Dict] = None
) -> None:
    base, ext = os.path.splitext(file_path)
    for i, image in enumerate(images):
        current_path = file_path
        counter = 1
        while os.path.exists(current_path):
            current_path = f"{base}_{counter}{ext}"
            counter += 1
        export_image(image, current_path, metadata[i] if metadata else None)



from PySide6.QtCore import Slot, QTimer, QPropertyAnimation
from PySide6.QtWidgets import QSpacerItem, QSizePolicy
from PySide6.QtCore import Qt

from airunner.enums import SignalCode, LLMActionType, ModelType, ModelStatus
from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.llm.templates.chat_prompt_ui import Ui_chat_prompt
from airunner.widgets.llm.message_widget import MessageWidget


class ChatPromptWidget(BaseWidget):
    widget_class_ = Ui_chat_prompt

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.scroll_bar = None
        self.conversation = None
        self.is_modal = True
        self.generating = False
        self.prefix = ""
        self.prompt = ""
        self.suffix = ""
        self.conversation_history = []
        self.spacer = None
        self.promptKeyPressEvent = None
        self.originalKeyPressEvent = None
        self.action_menu_displayed = None
        self.action_menu_displayed = None
        self.messages_spacer = None
        self.chat_loaded = False
        self.conversation_id = None

        self.ui.action.blockSignals(True)
        self.ui.action.addItem("Auto")
        self.ui.action.addItem("Chat")
        self.ui.action.addItem("Image")
        self.ui.action.addItem("RAG")
        action = LLMActionType[self.action]
        if action is LLMActionType.APPLICATION_COMMAND:
            self.ui.action.setCurrentIndex(0)
        elif action is LLMActionType.CHAT:
            self.ui.action.setCurrentIndex(1)
        elif action is LLMActionType.GENERATE_IMAGE:
            self.ui.action.setCurrentIndex(2)
        elif action is LLMActionType.PERFORM_RAG_SEARCH:
            self.ui.action.setCurrentIndex(3)
        self.ui.action.blockSignals(False)
        self.originalKeyPressEvent = None
        self.originalKeyPressEvent = self.ui.prompt.keyPressEvent
        self.register(SignalCode.AUDIO_PROCESSOR_RESPONSE_SIGNAL, self.on_hear_signal)
        self.register(SignalCode.SET_CONVERSATION, self.on_set_conversation)
        self.register(SignalCode.MODEL_STATUS_CHANGED_SIGNAL, self.on_model_status_changed_signal)
        self.register(SignalCode.CHATBOT_CHANGED, self.on_chatbot_changed)
        self.register(SignalCode.CONVERSATION_DELETED, self.on_conversation_deleted)
        self.held_message = None
        self._disabled = False
        self.scroll_animation = None

    @Slot(str)
    def handle_token_signal(self, val: str):
        if val != "[END]":
            text = self.ui.conversation.toPlainText()
            text += val
            self.ui.conversation.setPlainText(text)
        else:
            self.stop_progress_bar()
            self.generating = False
            self.enable_send_button()

    def on_model_status_changed_signal(self, data):
        if data["model"] == ModelType.LLM:
            self.chat_loaded = data["status"] is ModelStatus.LOADED

        if not self.chat_loaded:
            self.disable_send_button()
        else:
            self.enable_send_button()

    def on_chatbot_changed(self):
        self._clear_conversation()

    def on_conversation_deleted(self, data):
        if self.conversation_id == data["conversation_id"]:
            self.conversation_id = None
            self._clear_conversation()

    def on_set_conversation(self, message):
        self._clear_conversation_widgets()
        if len(message["messages"]) > 0:
            self.conversation_id = message["messages"][0]["conversation_id"]
        QTimer.singleShot(0, lambda: self._set_conversation_widgets(message["messages"]))

    def _set_conversation_widgets(self, messages):
        for message in messages:
            self.add_message_to_conversation(
                name=message["name"],
                message=message["content"],
                is_bot=message["is_bot"],
                message_id=message["id"],
                first_message=True
            )
        QTimer.singleShot(100, self.scroll_to_bottom)

    def on_hear_signal(self, data: dict):
        transcription = data["transcription"]
        self.ui.prompt.setPlainText(transcription)
        self.ui.send_button.click()

    def on_add_to_conversation_signal(self, name, text, is_bot):
        self.add_message_to_conversation(name=name, message=text, is_bot=is_bot)

    def on_add_bot_message_to_conversation(self, data: dict):
        try:
            name = data["name"]
            message = data["message"]
            is_first_message = data["is_first_message"]
            is_end_of_message = data["is_end_of_message"]
        except TypeError as e:
            self.logger.error("Error parsing data: "+str(e))
            self.enable_generate()
            return

        if is_first_message:
            self.stop_progress_bar()

        self.add_message_to_conversation(
            name=name,
            message=message,
            is_bot=True, 
            first_message=is_first_message
        )

        if is_end_of_message:
            self.enable_generate()

    def enable_generate(self):
        self.generating = False
        if self.held_message is not None:
            self.do_generate(prompt_override=self.held_message)
            self.held_message = None
        self.enable_send_button()

    @Slot()
    def action_button_clicked_clear_conversation(self):
        self._clear_conversation()

    def _clear_conversation(self):
        self.conversation_history = []
        self._clear_conversation_widgets()
        self._create_conversation()

    def _create_conversation(self):
        conversation = self.create_conversation()
        conversation_id = conversation.id
        self.emit_signal(SignalCode.LLM_CLEAR_HISTORY_SIGNAL, {
            "conversation_id": conversation_id
        })
        self.conversation_id = conversation_id

    def _clear_conversation_widgets(self):
        for widget in self.ui.scrollAreaWidgetContents.findChildren(MessageWidget):
            widget.deleteLater()
    
    @Slot(bool)
    def action_button_clicked_send(self):
        self.do_generate()

    def interrupt_button_clicked(self):
        self.emit_signal(SignalCode.INTERRUPT_PROCESS_SIGNAL)
        self.stop_progress_bar()
        self.generating = False
        self.enable_send_button()

    @property
    def action(self) -> str:
        return self.llm_generator_settings.action

    def do_generate(self, image_override=None, prompt_override=None, callback=None, generator_name="causallm"):
        if self.conversation_id is None:
            self._create_conversation()
        prompt = self.prompt if (prompt_override is None or prompt_override == "") else prompt_override
        if prompt is None or prompt == "":
            self.logger.warning("Prompt is empty")
            return

        if self.generating:
            if self.held_message is None:
                self.held_message = prompt
                self.disable_send_button()
                self.interrupt_button_clicked()
            return
        self.generating = True

        widget = self.add_message_to_conversation(
            name=self.user.username,
            message=self.prompt,
            is_bot=False
        )
        if widget is not None:
            QTimer.singleShot(100, widget.set_content_size)
            QTimer.singleShot(100, self.scroll_to_bottom)

        self.clear_prompt()
        self.start_progress_bar()
        self.emit_signal(
            SignalCode.LLM_TEXT_GENERATE_REQUEST_SIGNAL,
            {
                "llm_request": True,
                "request_data": {
                    "action": self.action,
                    "prompt": prompt,
                }
            }
        )

    def on_token_signal(self, val):
        self.handle_token_signal(val)

    registered = False

    def showEvent(self, event):
        super().showEvent(event)
        if not self.registered:
            self.register(SignalCode.LLM_TOKEN_SIGNAL, self.on_token_signal)
            self.register(SignalCode.LLM_TEXT_STREAMED_SIGNAL, self.on_add_bot_message_to_conversation)
            self.register(SignalCode.MODEL_STATUS_CHANGED_SIGNAL, self.on_model_status_changed)
            self.registered = True

        # handle return pressed on QPlainTextEdit
        # there is no returnPressed signal for QPlainTextEdit
        # so we have to use the keyPressEvent
        self.promptKeyPressEvent = self.ui.prompt.keyPressEvent

        # Override the method
        self.ui.prompt.keyPressEvent = self.handle_key_press

        if not self.chat_loaded:
            self.disable_send_button()

        self.ui.conversation.hide()
        self.ui.chat_container.show()

    def on_model_status_changed(self, data):
        if data["model"] == ModelType.LLM:
            if data["status"] is ModelStatus.LOADED:
                self.enable_send_button()
            else:
                self.disable_send_button()

    def llm_action_changed(self, val: str):
        if val == "Chat":
            llm_action_value = LLMActionType.CHAT
        elif val == "Image":
            llm_action_value = LLMActionType.GENERATE_IMAGE
        elif val == "RAG":
            llm_action_value = LLMActionType.PERFORM_RAG_SEARCH
        else:
            llm_action_value = LLMActionType.APPLICATION_COMMAND
        self.update_llm_generator_settings("action", llm_action_value.name)

    def prompt_text_changed(self):
        self.prompt = self.ui.prompt.toPlainText()

    def clear_prompt(self):
        self.ui.prompt.setPlainText("")

    def start_progress_bar(self):
        self.ui.progressBar.setRange(0, 0)
        self.ui.progressBar.setValue(0)

    def stop_progress_bar(self):
        self.ui.progressBar.setRange(0, 1)
        self.ui.progressBar.setValue(1)
        self.ui.progressBar.reset()

    def disable_send_button(self):
        self.ui.send_button.setEnabled(False)
        self._disabled = True

    def enable_send_button(self):
        self.ui.send_button.setEnabled(True)
        self._disabled = False

    def response_text_changed(self):
        pass

    def handle_key_press(self, event):
        if event.key() == Qt.Key.Key_Return:
            if not self._disabled and event.modifiers() != Qt.KeyboardModifier.ShiftModifier:
                self.do_generate()
                return
        # Call the original method
        if self.originalKeyPressEvent is not None and self.originalKeyPressEvent != self.handle_key_press:
            self.originalKeyPressEvent(event)

    def hide_action_menu(self):
        self.action_menu_displayed = False
        self.ui.action_menu.hide()

    def display_action_menu(self):
        self.action_menu_displayed = True
        self.ui.action_menu.show()

    def insert_newline(self):
        self.ui.prompt.insertPlainText("\n")
    
    def describe_image(self, image, callback):
        self.do_generate(
            image_override=image, 
            prompt_override="What is in this picture?",
            callback=callback,
            generator_name="visualqa"
        )

    def add_message_to_conversation(
        self,
        name,
        message,
        is_bot, 
        first_message=True,
        message_id=None
    ):
        if not first_message:
            # get the last widget from the scrollAreaWidgetContents.layout()
            # and append the message to it. must be a MessageWidget object
            # must start at the end of the layout and work backwards
            for i in range(self.ui.scrollAreaWidgetContents.layout().count() - 1, -1, -1):
                current_widget = self.ui.scrollAreaWidgetContents.layout().itemAt(i).widget()
                if isinstance(current_widget, MessageWidget):
                    if current_widget.is_bot:
                        if message != "":
                            current_widget.update_message(message)
                        self.scroll_to_bottom()
                        return
                    break

        self.remove_spacer()

        widget = None
        if message != "":
            widget = MessageWidget(
                name=name,
                message=message,
                is_bot=is_bot,
                message_id=message_id,
                conversation_id=self.conversation_id
            )
            self.ui.scrollAreaWidgetContents.layout().addWidget(widget)

        self.add_spacer()

        # automatically scroll to the bottom of the scrollAreaWidgetContents
        self.scroll_to_bottom()

        return widget

    def remove_spacer(self):
        if self.spacer is not None:
            self.ui.scrollAreaWidgetContents.layout().removeItem(self.spacer)

    def add_spacer(self):
        if self.spacer is None:
            self.spacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.ui.scrollAreaWidgetContents.layout().addItem(self.spacer)

    def message_type_text_changed(self, val):
        self.update_llm_generator_settings("message_type", val)

    def action_button_clicked_generate_characters(self):
        pass

    def scroll_to_bottom(self):
        if self.scroll_bar is None:
            self.scroll_bar = self.ui.chat_container.verticalScrollBar()

        if self.scroll_animation is None:
            self.scroll_animation = QPropertyAnimation(self.scroll_bar, b"value")
            self.scroll_animation.setDuration(500)

        # Stop any ongoing animation
        if self.scroll_animation and self.scroll_animation.state() == QPropertyAnimation.State.Running:
            self.scroll_animation.stop()

        self.scroll_animation.setStartValue(self.scroll_bar.value())
        self.scroll_animation.setEndValue(self.scroll_bar.maximum())
        self.scroll_animation.start()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'input_image.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QGraphicsView, QGridLayout, QHBoxLayout,
    QPushButton, QScrollArea, QSizePolicy, QSpacerItem,
    QWidget)

from airunner.widgets.controlnet.controlnet_settings_widget import ControlnetSettingsWidget
from airunner.widgets.slider.slider_widget import SliderWidget
from airunner.widgets.switch_widget.switch_widget import SwitchWidget
import airunner.resources_dark_rc
import airunner.resources_light_rc

class Ui_input_image(object):
    def setupUi(self, input_image):
        if not input_image.objectName():
            input_image.setObjectName(u"input_image")
        input_image.resize(581, 568)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(input_image.sizePolicy().hasHeightForWidth())
        input_image.setSizePolicy(sizePolicy)
        self.gridLayout_3 = QGridLayout(input_image)
        self.gridLayout_3.setSpacing(0)
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.gridLayout_3.setContentsMargins(0, 0, 0, 0)
        self.scrollArea = QScrollArea(input_image)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setWidgetResizable(True)
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setObjectName(u"scrollAreaWidgetContents")
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 565, 512))
        self.gridLayout_2 = QGridLayout(self.scrollAreaWidgetContents)
        self.gridLayout_2.setSpacing(0)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)
        self.image_container = QGraphicsView(self.scrollAreaWidgetContents)
        self.image_container.setObjectName(u"image_container")
        sizePolicy.setHeightForWidth(self.image_container.sizePolicy().hasHeightForWidth())
        self.image_container.setSizePolicy(sizePolicy)
        self.image_container.setMinimumSize(QSize(512, 512))

        self.gridLayout_2.addWidget(self.image_container, 0, 0, 1, 1)

        self.scrollArea.setWidget(self.scrollAreaWidgetContents)

        self.gridLayout_3.addWidget(self.scrollArea, 0, 0, 1, 1)

        self.verticalWidget = QWidget(input_image)
        self.verticalWidget.setObjectName(u"verticalWidget")
        self.gridLayout = QGridLayout(self.verticalWidget)
        self.gridLayout.setSpacing(10)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(10, 10, 10, 10)
        self.strength_slider_widget = SliderWidget(self.verticalWidget)
        self.strength_slider_widget.setObjectName(u"strength_slider_widget")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Minimum)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.strength_slider_widget.sizePolicy().hasHeightForWidth())
        self.strength_slider_widget.setSizePolicy(sizePolicy1)
        self.strength_slider_widget.setProperty("slider_maximum", 100)
        self.strength_slider_widget.setProperty("current_value", 0)
        self.strength_slider_widget.setProperty("spinbox_maximum", 1.000000000000000)
        self.strength_slider_widget.setProperty("display_as_float", True)
        self.strength_slider_widget.setProperty("spinbox_single_step", 0.010000000000000)
        self.strength_slider_widget.setProperty("spinbox_page_step", 0.100000000000000)

        self.gridLayout.addWidget(self.strength_slider_widget, 0, 0, 1, 1)

        self.mask_blur_slider_widget = SliderWidget(self.verticalWidget)
        self.mask_blur_slider_widget.setObjectName(u"mask_blur_slider_widget")
        sizePolicy1.setHeightForWidth(self.mask_blur_slider_widget.sizePolicy().hasHeightForWidth())
        self.mask_blur_slider_widget.setSizePolicy(sizePolicy1)
        self.mask_blur_slider_widget.setProperty("slider_maximum", 100)
        self.mask_blur_slider_widget.setProperty("current_value", 0)
        self.mask_blur_slider_widget.setProperty("spinbox_maximum", 1.000000000000000)
        self.mask_blur_slider_widget.setProperty("display_as_float", True)
        self.mask_blur_slider_widget.setProperty("spinbox_single_step", 0.010000000000000)
        self.mask_blur_slider_widget.setProperty("spinbox_page_step", 0.100000000000000)

        self.gridLayout.addWidget(self.mask_blur_slider_widget, 1, 0, 1, 1)

        self.controlnet_settings = ControlnetSettingsWidget(self.verticalWidget)
        self.controlnet_settings.setObjectName(u"controlnet_settings")

        self.gridLayout.addWidget(self.controlnet_settings, 2, 0, 1, 1)


        self.gridLayout_3.addWidget(self.verticalWidget, 1, 0, 1, 1)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSpacing(10)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(10, -1, 10, 10)
        self.link_to_grid_image_button = QPushButton(input_image)
        self.link_to_grid_image_button.setObjectName(u"link_to_grid_image_button")
        icon = QIcon(QIcon.fromTheme(u"insert-link"))
        self.link_to_grid_image_button.setIcon(icon)
        self.link_to_grid_image_button.setCheckable(True)

        self.horizontalLayout.addWidget(self.link_to_grid_image_button)

        self.lock_input_image_button = QPushButton(input_image)
        self.lock_input_image_button.setObjectName(u"lock_input_image_button")
        self.lock_input_image_button.setMinimumSize(QSize(24, 24))
        self.lock_input_image_button.setMaximumSize(QSize(24, 24))
        icon1 = QIcon(QIcon.fromTheme(u"emblem-readonly"))
        self.lock_input_image_button.setIcon(icon1)
        self.lock_input_image_button.setCheckable(True)

        self.horizontalLayout.addWidget(self.lock_input_image_button)

        self.pushButton_3 = QPushButton(input_image)
        self.pushButton_3.setObjectName(u"pushButton_3")
        icon2 = QIcon(QIcon.fromTheme(u"view-refresh"))
        self.pushButton_3.setIcon(icon2)

        self.horizontalLayout.addWidget(self.pushButton_3)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.EnableSwitch = SwitchWidget(input_image)
        self.EnableSwitch.setObjectName(u"EnableSwitch")
        self.EnableSwitch.setMinimumSize(QSize(45, 20))
        self.EnableSwitch.setMaximumSize(QSize(45, 20))

        self.horizontalLayout.addWidget(self.EnableSwitch)

        self.import_button = QPushButton(input_image)
        self.import_button.setObjectName(u"import_button")
        icon3 = QIcon(QIcon.fromTheme(u"folder"))
        self.import_button.setIcon(icon3)

        self.horizontalLayout.addWidget(self.import_button)

        self.delete_button = QPushButton(input_image)
        self.delete_button.setObjectName(u"delete_button")
        icon4 = QIcon(QIcon.fromTheme(u"user-trash"))
        self.delete_button.setIcon(icon4)

        self.horizontalLayout.addWidget(self.delete_button)


        self.gridLayout_3.addLayout(self.horizontalLayout, 2, 0, 1, 1)


        self.retranslateUi(input_image)
        self.import_button.clicked.connect(input_image.import_clicked)
        self.delete_button.clicked.connect(input_image.delete_clicked)
        self.link_to_grid_image_button.toggled.connect(input_image.use_grid_image_as_input_toggled)
        self.lock_input_image_button.toggled.connect(input_image.lock_input_image)
        self.pushButton_3.clicked.connect(input_image.refresh_input_image_from_grid)

        QMetaObject.connectSlotsByName(input_image)
    # setupUi

    def retranslateUi(self, input_image):
        input_image.setWindowTitle(QCoreApplication.translate("input_image", u"Form", None))
        self.strength_slider_widget.setProperty("label_text", QCoreApplication.translate("input_image", u"Strength", None))
        self.strength_slider_widget.setProperty("settings_property", QCoreApplication.translate("input_image", u"generator_settings.strength", None))
        self.mask_blur_slider_widget.setProperty("label_text", QCoreApplication.translate("input_image", u"Mask Blur", None))
        self.mask_blur_slider_widget.setProperty("settings_property", QCoreApplication.translate("input_image", u"outpaint_settings.mask_blur", None))
#if QT_CONFIG(tooltip)
        self.link_to_grid_image_button.setToolTip(QCoreApplication.translate("input_image", u"Link to Grid image", None))
#endif // QT_CONFIG(tooltip)
        self.link_to_grid_image_button.setText("")
#if QT_CONFIG(tooltip)
        self.lock_input_image_button.setToolTip(QCoreApplication.translate("input_image", u"Lock current input image", None))
#endif // QT_CONFIG(tooltip)
        self.lock_input_image_button.setText("")
#if QT_CONFIG(tooltip)
        self.pushButton_3.setToolTip(QCoreApplication.translate("input_image", u"Refresh from grid image", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_3.setText("")
#if QT_CONFIG(tooltip)
        self.import_button.setToolTip(QCoreApplication.translate("input_image", u"Import", None))
#endif // QT_CONFIG(tooltip)
        self.import_button.setText("")
#if QT_CONFIG(tooltip)
        self.delete_button.setToolTip(QCoreApplication.translate("input_image", u"Delete", None))
#endif // QT_CONFIG(tooltip)
        self.delete_button.setText("")
    # retranslateUi




from PySide6.QtCore import QTimer, Slot

from airunner.data.models.settings_models import Lora
from airunner.enums import SignalCode
from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.lora.lora_trigger_word_widget import LoraTriggerWordWidget
from airunner.widgets.lora.templates.lora_ui import Ui_lora


class LoraWidget(BaseWidget):
    """
    This class represents a single lora.
    It is responsible for displaying the lora's name, trigger words,
    and active status.
    """
    widget_class_ = Ui_lora

    def __init__(self, *args, **kwargs):
        self.icons = [
            ("recycle-bin-line-icon", "delete_button"),
        ]
        self.current_lora:Lora = kwargs.pop("lora")
        super().__init__(*args, **kwargs)
        name = self.current_lora.name
        enabled = self.current_lora.enabled
        trigger_word = self.current_lora.trigger_word

        # Block signals for batch updates
        self.ui.enabledCheckbox.blockSignals(True)
        self.ui.trigger_word_edit.blockSignals(True)

        self.ui.enabledCheckbox.setText(name)
        self.ui.enabledCheckbox.setChecked(enabled)
        self.ui.trigger_word_edit.setText(trigger_word)

        # Unblock signals after batch updates
        self.ui.enabledCheckbox.blockSignals(False)
        self.ui.trigger_word_edit.blockSignals(False)

        # Defer the creation of trigger word widgets
        self.create_trigger_word_widgets(self.current_lora, defer=True)

        self.ui.scale_slider.setProperty("table_id", self.current_lora.id)

    def disable_lora_widget(self):
        self.ui.enabledCheckbox.setEnabled(False)
        self.ui.delete_button.setEnabled(False)
        self.ui.scale_slider.setEnabled(False)

    def enable_lora_widget(self):
        self.ui.enabledCheckbox.setEnabled(True)
        self.ui.delete_button.setEnabled(True)
        self.ui.scale_slider.setEnabled(True)

    @Slot(bool)
    def action_toggled_lora_enabled(self, val):
        self.current_lora.enabled = val
        self.ui.enabledCheckbox.blockSignals(True)
        self.ui.enabledCheckbox.setChecked(val)
        self.ui.enabledCheckbox.blockSignals(False)
        self.update_lora(self.current_lora)
        self.emit_signal(SignalCode.LORA_STATUS_CHANGED)

    @Slot(str)
    def action_text_changed_trigger_word(self, val):
        self.current_lora.trigger_word = val
        self.update_lora(self.current_lora)

    @Slot()
    def action_clicked_button_deleted(self):
        self.emit_signal(
            SignalCode.LORA_DELETE_SIGNAL,
            {
                "lora_widget": self
            }
        )

    def create_trigger_word_widgets(self, lora, defer=False):
        if defer:
            # Defer the creation of trigger word widgets
            QTimer.singleShot(0, lambda: self._create_trigger_word_widgets(lora))
        else:
            self._create_trigger_word_widgets(lora)

    def _create_trigger_word_widgets(self, lora):
        for i in reversed(range(self.ui.lora_container.layout().count())):
            widget = self.ui.lora_container.layout().itemAt(i).widget()
            if isinstance(widget, LoraTriggerWordWidget):
                widget.deleteLater()
        for word in lora.trigger_word.split(","):
            if word.strip() == "":
                continue
            widget = LoraTriggerWordWidget(trigger_word=word)
            self.ui.lora_container.layout().addWidget(widget)

    def action_changed_trigger_words(self, val):
        self.current_lora.trigger_word = val
        self.update_lora(self.current_lora)
        self.create_trigger_word_widgets(self.current_lora)



from typing import Any

from airunner.widgets.slider.slider_widget import SliderWidget


class FilterSliderWidget(SliderWidget):
    def __init__(self, *args, session, filter_value, preview_filter, **kwargs):
        self._filter_value = filter_value
        self.preview_filter = preview_filter
        super().__init__(*args, **kwargs)

    def init(self, **kwargs):
        if kwargs is not None and kwargs != {}:
            super().init(**kwargs)
        return

    def get_settings_value(self, settings_property):
        value_type = self._filter_value.value_type
        if value_type == "float":
            return float(self._filter_value.value)
        elif value_type == "int":
            return int(self._filter_value.value)
        else:
            return self._filter_value.value

    def set_settings_value(self, settings_property: str, val: Any):
        self._filter_value.value = val
        self.session.commit()
        self.preview_filter()



from alembic import op
import sqlalchemy as sa

def column_exists(table_name, column_name):
    conn = op.get_bind()
    inspector = sa.inspect(conn)
    columns = [col['name'] for col in inspector.get_columns(table_name)]
    return column_name in columns



import os
from pathlib import Path

from airunner.settings import DARK_THEME_NAME, LIGHT_THEME_NAME


class StylesMixin:
    """
    Dependent on the SettingsMixin being used in the same class
    """
    def set_stylesheet(self):
        """
        Sets the stylesheet for the application based on the current theme
        """
        theme_name = DARK_THEME_NAME if self.application_settings.dark_mode_enabled else LIGHT_THEME_NAME
        base_dir = Path(os.path.dirname(os.path.realpath(__file__)))
        stylesheet_path = base_dir / "styles" / theme_name / "styles.qss"

        self.setStyleSheet(
            stylesheet_path.read_text()
            if self.application_settings.override_system_theme
            else ""
        )



from airunner.windows.base_window import BaseWindow


class ThemeEditor(BaseWindow):
    template_name = "theme_editor"
    window_title = "Theme Editor"

    def initialize_window(self):
        pass


import os

from PIL import Image
from PySide6.QtCore import Slot
from PySide6.QtWidgets import QFileDialog
from PIL.ImageQt import ImageQt
from PySide6.QtGui import QPixmap, QImage, Qt, QPen
from PySide6.QtWidgets import QGraphicsScene

from airunner.settings import VALID_IMAGE_FILES
from airunner.utils.image.convert_binary_to_image import convert_binary_to_image
from airunner.utils.image.convert_image_to_binary import convert_image_to_binary
from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.canvas.templates.input_image_ui import Ui_input_image


class InputImage(BaseWidget):
    widget_class_ = Ui_input_image

    def __init__(self, *args, **kwargs):
        self.settings_key = kwargs.pop("settings_key")
        self.use_generated_image = kwargs.pop("use_generated_image", False)
        self._import_path = ""
        super().__init__(*args, **kwargs)

    @property
    def current_settings(self):
        settings = None
        if self.settings_key == "controlnet_settings":
            settings = self.controlnet_settings
        elif self.settings_key == "image_to_image_settings":
            settings = self.image_to_image_settings
        elif self.settings_key == "outpaint_settings":
            settings = self.outpaint_settings
        elif self.settings_key == "drawing_pad_settings":
            settings = self.drawing_pad_settings

        if not settings:
            raise ValueError(f"Settings not found for key: {self.settings_key}")

        return settings

    def on_mask_generator_worker_response_signal(self):
        if self.settings_key == "outpaint_settings":
            self.load_image_from_settings()

    def update_current_settings(self, key, value):
        if self.settings_key == "controlnet_settings":
            self.update_controlnet_settings(key, value)
        elif self.settings_key == "image_to_image_settings":
            self.update_image_to_image_settings(key, value)
        elif self.settings_key == "outpaint_settings":
            self.update_outpaint_settings(key, value)
        elif self.settings_key == "drawing_pad_settings":
            self.update_drawing_pad_settings(key, value)

    def showEvent(self, event):
        super().showEvent(event)
        if self.settings_key == "controlnet_settings":
            self.ui.strength_slider_widget.hide()
            self.ui.controlnet_settings.show()
        else:
            self.ui.strength_slider_widget.show()
            self.ui.controlnet_settings.hide()

        if self.settings_key == "outpaint_settings":
            self.ui.strength_slider_widget.setProperty("settings_property", 'outpaint_settings.strength')
            self.ui.mask_blur_slider_widget.show()
        else:
            self.ui.mask_blur_slider_widget.hide()

        self.ui.EnableSwitch.toggled.connect(self.enabled_toggled)
        if self.settings_key == "outpaint_settings":
            self.ui.import_button.hide()
            self.ui.link_to_grid_image_button.hide()
            self.ui.link_to_grid_image_button.hide()
            self.ui.lock_input_image_button.hide()
            self.ui.EnableSwitch.blockSignals(True)
            self.ui.EnableSwitch.checked = self.current_settings.enabled
            self.ui.EnableSwitch.setChecked(self.current_settings.enabled)
            self.ui.EnableSwitch.dPtr.animate(self.current_settings.enabled)
            self.ui.EnableSwitch.blockSignals(False)
        else:
            self.ui.EnableSwitch.blockSignals(True)
            self.ui.link_to_grid_image_button.blockSignals(True)
            self.ui.lock_input_image_button.blockSignals(True)
            self.ui.EnableSwitch.checked = self.current_settings.enabled
            self.ui.EnableSwitch.setChecked(self.current_settings.enabled)
            self.ui.EnableSwitch.dPtr.animate(self.current_settings.enabled)
            self.ui.link_to_grid_image_button.setChecked(
                self.current_settings.use_grid_image_as_input
            )
            self.ui.lock_input_image_button.setChecked(
                self.current_settings.lock_input_image or False  # Provide a default value
            )
            self.ui.EnableSwitch.blockSignals(False)
            self.ui.link_to_grid_image_button.blockSignals(False)
            self.ui.lock_input_image_button.blockSignals(False)

            if self.current_settings.use_grid_image_as_input:
                self.load_image_from_grid()
                return
        self.load_image_from_settings()

    @Slot(bool)
    def enabled_toggled(self, val):
        self.update_current_settings("enabled", val)

    @Slot(bool)
    def lock_input_image(self, val):
        self.update_current_settings("lock_input_image", val)

    @Slot(bool)
    def refresh_input_image_from_grid(self):
        self.load_image_from_grid(forced=True)

    @Slot(bool)
    def use_grid_image_as_input_toggled(self, val):
        self.update_current_settings("use_grid_image_as_input", val)
        if val is True:
            self.load_image_from_grid()

    @Slot()
    def import_clicked(self):
        self.import_image()

    @Slot()
    def delete_clicked(self):
        self.delete_image()

    def import_image(self):
        self._import_path, _ = QFileDialog.getOpenFileName(
            self.window(),
            "Open Image",
            self._import_path,
            f"Image Files ({' '.join(VALID_IMAGE_FILES)})"
        )
        if self._import_path == "":
            return
        self.load_image(
            os.path.abspath(self._import_path)
        )

    def load_image(self, file_path: str):
        image = Image.open(file_path)
        self.load_image_from_object(image)
        if image is not None:
            self.update_current_settings("image", convert_image_to_binary(image))

    def load_image_from_grid(self, forced=False):
        if self.settings_key == "outpaint_settings":
            return
        if not forced and not self.current_settings.use_grid_image_as_input:
            return
        if not forced and self.current_settings.lock_input_image:
            return
        self.update_current_settings("image", self.drawing_pad_settings.image)
        self.load_image_from_settings()

    def load_image_from_settings(self):
        if self.use_generated_image:
            image = self.current_settings.generated_image
        else:
            image = self.current_settings.image

        if self.settings_key == "outpaint_settings":
            image = self.drawing_pad_settings.mask

        if image is not None:
            image = convert_binary_to_image(image)
            if image is not None:
                self.load_image_from_object(image)
                return
        self.ui.image_container.setScene(None)

    def load_image_from_object(self, image: Image):
        if image is None:
            self.logger.warning("Image is None, unable to add to scene")
            return

        # Resize the image to maintain aspect ratio, but not exceed 512x512
        max_size = (512, 512)
        image.thumbnail(max_size, Image.Resampling.LANCZOS)

        # Convert PIL image to QImage
        qimage = ImageQt(image)
        qpixmap = QPixmap.fromImage(QImage(qimage))

        # Create a QGraphicsScene and clear it
        scene = QGraphicsScene()
        scene.clear()  # Clear the scene before adding new items
        scene.addPixmap(qpixmap)

        # Set scene width and height
        scene.setSceneRect(0, 0, qpixmap.width(), qpixmap.height())

        # Set the QGraphicsScene to the QGraphicsView
        self.ui.image_container.setScene(scene)

        # Set the alignment to top-left corner
        self.ui.image_container.setAlignment(Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignTop)

        # Draw a red border around the image
        pen = QPen(Qt.GlobalColor.red)
        pen.setWidth(3)  # Set the width of the border
        scene.addRect(0, 0, qpixmap.width(), qpixmap.height(), pen)

    def delete_image(self):
        if self.settings_key == "outpaint_settings":
            self.update_drawing_pad_settings("mask", None)
        else:
            self.update_current_settings("image", None)
        self.ui.image_container.setScene(None)



from airunner.windows.setup_wizard.user_agreement.agreement_page import AgreementPage
from airunner.windows.setup_wizard.model_setup.stable_diffusion_setup.templates.stable_diffusion_license_ui import Ui_stable_diffusion_license


class StableDiffusionLicense(AgreementPage):
    class_name_ = Ui_stable_diffusion_license
    setting_key = "stable_diffusion_agreement_checked"





"""Move bot_mood from Chatbot to Conversation

Revision ID: f403ff7468e8
Revises: 536e18463461
Create Date: 2024-10-24 01:43:30.378281

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.exc import OperationalError

# revision identifiers, used by Alembic.
revision: str = 'f403ff7468e8'
down_revision: Union[str, None] = '536e18463461'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None

def upgrade():
    bind = op.get_bind()
    inspector = Inspector.from_engine(bind)
    columns = [col['name'] for col in inspector.get_columns('conversations')]

    if 'bot_mood' not in columns:
        op.add_column('conversations', sa.Column('bot_mood', sa.Text(), nullable=True))

    try:
        op.drop_column('chatbots', 'bot_mood')
    except OperationalError as e:
        print(f"Error dropping column 'bot_mood' from 'chatbots': {e}")

def downgrade():
    op.add_column('chatbots', sa.Column('bot_mood', sa.Text(), nullable=True))
    op.drop_column('conversations', 'bot_mood')



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'default.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QComboBox, QGridLayout,
    QHBoxLayout, QLineEdit, QScrollArea, QSizePolicy,
    QVBoxLayout, QWidget)

class Ui_default_model_widget(object):
    def setupUi(self, default_model_widget):
        if not default_model_widget.objectName():
            default_model_widget.setObjectName(u"default_model_widget")
        default_model_widget.resize(496, 434)
        self.gridLayout = QGridLayout(default_model_widget)
        self.gridLayout.setObjectName(u"gridLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.comboBox = QComboBox(default_model_widget)
        self.comboBox.setObjectName(u"comboBox")

        self.horizontalLayout.addWidget(self.comboBox)

        self.toggle_all = QCheckBox(default_model_widget)
        self.toggle_all.setObjectName(u"toggle_all")
        font = QFont()
        font.setPointSize(8)
        self.toggle_all.setFont(font)
        self.toggle_all.setTristate(True)

        self.horizontalLayout.addWidget(self.toggle_all)


        self.gridLayout.addLayout(self.horizontalLayout, 0, 0, 1, 1)

        self.default_scroll_area = QScrollArea(default_model_widget)
        self.default_scroll_area.setObjectName(u"default_scroll_area")
        self.default_scroll_area.setWidgetResizable(True)
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setObjectName(u"scrollAreaWidgetContents")
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 476, 350))
        self.verticalLayout = QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.default_scroll_area.setWidget(self.scrollAreaWidgetContents)

        self.gridLayout.addWidget(self.default_scroll_area, 2, 0, 1, 1)

        self.lineEdit = QLineEdit(default_model_widget)
        self.lineEdit.setObjectName(u"lineEdit")

        self.gridLayout.addWidget(self.lineEdit, 1, 0, 1, 1)


        self.retranslateUi(default_model_widget)
        self.comboBox.currentTextChanged.connect(default_model_widget.mode_type_changed)
        self.lineEdit.textEdited.connect(default_model_widget.search_text_changed)
        self.toggle_all.stateChanged.connect(default_model_widget.toggle_all_state_change)

        QMetaObject.connectSlotsByName(default_model_widget)
    # setupUi

    def retranslateUi(self, default_model_widget):
        default_model_widget.setWindowTitle(QCoreApplication.translate("default_model_widget", u"Form", None))
        self.toggle_all.setText(QCoreApplication.translate("default_model_widget", u"Toggle all", None))
        self.lineEdit.setPlaceholderText(QCoreApplication.translate("default_model_widget", u"Search models", None))
    # retranslateUi




"""Change image and mask columns to BLOB

Revision ID: 26a0d29a3af3
Revises: f403ff7468e8
Create Date: 2024-10-24 12:31:37.706170

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision: str = '26a0d29a3af3'
down_revision: Union[str, None] = 'f403ff7468e8'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade():
    try:
        op.alter_column('controlnet_settings', 'image', type_=sa.LargeBinary, existing_type=sa.String, nullable=True)
        op.alter_column('controlnet_settings', 'generated_image', type_=sa.LargeBinary, existing_type=sa.String, nullable=True)
    except sa.exc.OperationalError:
        pass

    try:
        op.alter_column('image_to_image_settings', 'image', type_=sa.LargeBinary, existing_type=sa.String, nullable=True)
    except sa.exc.OperationalError:
        pass

    try:
        op.alter_column('outpaint_settings', 'image', type_=sa.LargeBinary, existing_type=sa.String, nullable=True)
    except sa.exc.OperationalError:
        pass

    try:
        op.alter_column('drawing_pad_settings', 'image', type_=sa.LargeBinary, existing_type=sa.String, nullable=True)
        op.alter_column('drawing_pad_settings', 'mask', type_=sa.LargeBinary, existing_type=sa.String, nullable=True)
    except sa.exc.OperationalError:
        pass


def downgrade():
    try:
        op.alter_column('controlnet_settings', 'image', type_=sa.String, existing_type=sa.LargeBinary, nullable=True)
        op.alter_column('controlnet_settings', 'generated_image', type_=sa.String, existing_type=sa.LargeBinary, nullable=True)
    except sa.exc.OperationalError:
        pass

    try:
        op.alter_column('image_to_image_settings', 'image', type_=sa.String, existing_type=sa.LargeBinary, nullable=True)
    except sa.exc.OperationalError:
        pass

    try:
        op.alter_column('outpaint_settings', 'image', type_=sa.String, existing_type=sa.LargeBinary, nullable=True)
    except sa.exc.OperationalError:
        pass

    try:
        op.alter_column('drawing_pad_settings', 'image', type_=sa.String, existing_type=sa.LargeBinary, nullable=True)
        op.alter_column('drawing_pad_settings', 'mask', type_=sa.String, existing_type=sa.LargeBinary, nullable=True)
    except sa.exc.OperationalError:
        pass



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'generated_image.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QGridLayout, QHBoxLayout,
    QLabel, QPushButton, QSizePolicy, QSpacerItem,
    QWidget)

class Ui_Form(object):
    def setupUi(self, Form):
        if not Form.objectName():
            Form.setObjectName(u"Form")
        Form.resize(283, 328)
        self.gridLayout = QGridLayout(Form)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setHorizontalSpacing(0)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalSpacer = QSpacerItem(17, 0, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer, 5, 0, 1, 1)

        self.image_container = QLabel(Form)
        self.image_container.setObjectName(u"image_container")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.image_container.sizePolicy().hasHeightForWidth())
        self.image_container.setSizePolicy(sizePolicy)
        self.image_container.setMinimumSize(QSize(256, 256))
        self.image_container.setMaximumSize(QSize(256, 256))

        self.gridLayout.addWidget(self.image_container, 0, 0, 1, 1)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.to_canvas_button = QPushButton(Form)
        self.to_canvas_button.setObjectName(u"to_canvas_button")

        self.horizontalLayout.addWidget(self.to_canvas_button)

        self.export_button = QPushButton(Form)
        self.export_button.setObjectName(u"export_button")

        self.horizontalLayout.addWidget(self.export_button)

        self.delete_button = QPushButton(Form)
        self.delete_button.setObjectName(u"delete_button")

        self.horizontalLayout.addWidget(self.delete_button)


        self.gridLayout.addLayout(self.horizontalLayout, 1, 0, 1, 1)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.slot_combobox = QComboBox(Form)
        self.slot_combobox.setObjectName(u"slot_combobox")

        self.horizontalLayout_2.addWidget(self.slot_combobox)

        self.to_slot_button = QPushButton(Form)
        self.to_slot_button.setObjectName(u"to_slot_button")

        self.horizontalLayout_2.addWidget(self.to_slot_button)


        self.gridLayout.addLayout(self.horizontalLayout_2, 3, 0, 1, 1)


        self.retranslateUi(Form)

        QMetaObject.connectSlotsByName(Form)
    # setupUi

    def retranslateUi(self, Form):
        Form.setWindowTitle(QCoreApplication.translate("Form", u"Form", None))
        self.image_container.setText("")
        self.to_canvas_button.setText(QCoreApplication.translate("Form", u"Send to canvas", None))
        self.export_button.setText(QCoreApplication.translate("Form", u"Export", None))
        self.delete_button.setText(QCoreApplication.translate("Form", u"Delete", None))
        self.to_slot_button.setText(QCoreApplication.translate("Form", u"Send to slot", None))
    # retranslateUi




import os
from PySide6.QtWidgets import QFileDialog
from PySide6.QtCore import Slot

from airunner.settings import BASE_PATH, DEFAULT_PATH_SETTINGS
from airunner.utils.os.create_airunner_directory import create_airunner_paths
from airunner.windows.setup_wizard.base_wizard import BaseWizard
from airunner.windows.setup_wizard.path_settings.templates.path_settings_ui import Ui_PathSettings
from airunner.enums import SignalCode


class PathSettings(BaseWizard):
    class_name_ = Ui_PathSettings

    def __init__(self, *args):
        super(PathSettings, self).__init__(*args)

        self.ui.base_path.setText(BASE_PATH)

    @Slot()
    def browse_files(self):
        base_path = str(self.path_settings.base_path)
        file_dialog = QFileDialog()
        file_dialog.setFileMode(QFileDialog.Directory)
        file_dialog.setOption(QFileDialog.ShowDirsOnly)
        file_dialog.setDirectory(base_path)
        file_dialog.setModal(True)
        file_dialog.exec_()
        selected_path = file_dialog.selectedFiles()
        if selected_path:
            self.ui.base_path.setText(selected_path[0])
        base_path = self.ui.base_path.text()
        self.update_path_settings("base_path", base_path)
        for k, v in DEFAULT_PATH_SETTINGS.items():
                self.update_path_settings(k, os.path.expanduser(os.path.join(base_path, v)))




import threading

import torch
from PySide6.QtCore import QThread
from PySide6.QtCore import QObject, Signal, Slot

from airunner.enums import QueueType, SignalCode, ModelType, ModelAction
from airunner.mediator_mixin import MediatorMixin
from airunner.windows.main.settings_mixin import SettingsMixin
from airunner.workers.worker import Worker

torch.backends.cuda.matmul.allow_tf32 = True


class GenerateWorker(QObject):
    finished = Signal()
    error = Signal(str)

    def __init__(self, sd, message):
        super().__init__()
        self.sd = sd
        self.message = message

    @Slot()
    def run(self):
        try:
            self.sd.handle_generate_signal(self.message)
        except Exception as e:
            self.error.emit(str(e))
        finally:
            self.finished.emit()


class SDWorker(Worker):
    queue_type = QueueType.GET_LAST_ITEM

    def __init__(self):
        self.sd = None
        MediatorMixin.__init__(self)
        
        super().__init__(
            signals=(
                (SignalCode.RESET_APPLIED_MEMORY_SETTINGS, self.on_reset_applied_memory_settings),
                (SignalCode.SD_CANCEL_SIGNAL, self.on_sd_cancel_signal),
                (SignalCode.START_AUTO_IMAGE_GENERATION_SIGNAL, self.on_start_auto_image_generation_signal),
                (SignalCode.STOP_AUTO_IMAGE_GENERATION_SIGNAL, self.on_stop_auto_image_generation_signal),
                (SignalCode.DO_GENERATE_SIGNAL, self.on_do_generate_signal),
                (SignalCode.INTERRUPT_IMAGE_GENERATION_SIGNAL, self.on_interrupt_image_generation_signal),
                (SignalCode.CHANGE_SCHEDULER_SIGNAL, self.on_change_scheduler_signal),
                (SignalCode.MODEL_STATUS_CHANGED_SIGNAL, self.on_model_status_changed_signal),
                (SignalCode.SD_LOAD_SIGNAL, self.on_load_stablediffusion_signal),
                (SignalCode.SD_UNLOAD_SIGNAL, self.on_unload_stablediffusion_signal),
                (SignalCode.CONTROLNET_LOAD_SIGNAL, self.on_load_controlnet_signal),
                (SignalCode.CONTROLNET_UNLOAD_SIGNAL, self.on_unload_controlnet_signal),
                (SignalCode.LORA_UPDATE_SIGNAL, self.on_update_lora_signal),
                (SignalCode.EMBEDDING_UPDATE_SIGNAL, self.on_update_embeddings_signal),
                (SignalCode.EMBEDDING_DELETE_MISSING_SIGNAL, self.delete_missing_embeddings),
                (SignalCode.SD_STATE_CHANGED_SIGNAL, self.handle_sd_state_changed_signal),
                (SignalCode.SAFETY_CHECKER_LOAD_SIGNAL, self.on_load_safety_checker),
                (SignalCode.SAFETY_CHECKER_UNLOAD_SIGNAL, self.on_unload_safety_checker),
            )
        )
        self.__requested_action = ModelAction.NONE
        self._threads = []
        self._workers = []

    def handle_sd_state_changed_signal(self, _data=None):
        self.sd.controlnet_handle_sd_state_changed_signal()
        self.sd.scheduler_handle_sd_state_changed_signal()

    def on_load_safety_checker(self):
        if self.sd:
            thread = threading.Thread(target=self._load_safety_checker)
            thread.start()

    def on_unload_safety_checker(self):
        if self.sd:
            thread = threading.Thread(target=self._unload_safety_checker)
            thread.start()

    def scan_for_embeddings(self):
        if self.sd:
            self.sd.scan_for_embeddings()

    def delete_missing_embeddings(self, message):
        if self.sd:
            self.sd.delete_missing_embeddings(message)

    def get_embeddings(self, message):
        if self.sd:
            self.sd.get_embeddings(message)

    def on_update_lora_signal(self):
        thread = threading.Thread(target=self._reload_lora)
        thread.start()

    def _reload_lora(self):
        if self.sd:
            self.sd.reload_lora()

    def on_update_embeddings_signal(self):
        if self.sd:
            self.sd.reload_embeddings()

    def on_add_lora_signal(self, message):
        if self.sd:
            self.sd.on_add_lora_signal(message)

    def on_load_controlnet_signal(self, _data=None):
        if self.sd:
            thread = threading.Thread(target=self._load_controlnet)
            thread.start()

    def on_unload_controlnet_signal(self, _data=None):
        if self.sd:
            thread = threading.Thread(target=self._unload_controlnet)
            thread.start()

    def on_load_stablediffusion_signal(self, data:dict=None):
        if self.sd:
            thread = threading.Thread(target=self._load_sd, args=(data,))
            thread.start()

    def on_unload_stablediffusion_signal(self, data=None):
        if self.sd:
            thread = threading.Thread(target=self._unload_sd, args=(data,))
            thread.start()

    def _load_sd(self, data:dict=None):
        do_reload = data.get("do_reload", False)
        if do_reload:
            self.sd.reload()
        else:
            self.sd.load()
        if data:
            callback = data.get("callback", None)
            if callback is not None:
                callback(data)

    def _unload_sd(self, data:dict=None):
        self.sd.unload()
        if data:
            callback = data.get("callback", None)
            if callback is not None:
                callback(data)

    def _load_controlnet(self):
        self.sd.load_controlnet()

    def _unload_controlnet(self):
        self.sd.unload_controlnet()

    def _load_safety_checker(self):
        self.sd.load_safety_checker()

    def _unload_safety_checker(self):
        self.sd.unload_safety_checker()

    def on_tokenizer_load_signal(self, data: dict = None):
        if self.sd:
            self.sd.sd_load_tokenizer(data)

    def start_worker_thread(self):
        from airunner.handlers.stablediffusion.sd_handler import SDHandler
        self.sd = SDHandler()
        if self.application_settings.sd_enabled:
            self.sd.load()

    def handle_message(self, message):
        if self.sd:
            self.sd.run()

    def on_reset_applied_memory_settings(self, _data=None):
        if self.sd:
            self.sd.reset_applied_memory_settings()

    def on_sd_cancel_signal(self, _data=None):
        print("on_sd_cancel_signal")

    def on_start_auto_image_generation_signal(self, _data=None):
        # self.sd_mode = SDMode.DRAWING
        # self.generate()
        pass

    def on_stop_auto_image_generation_signal(self, _data=None):
        #self.sd_mode = SDMode.STANDARD
        pass

    def on_do_generate_signal(self, message: dict):
        if self.sd:
            thread = QThread()
            worker = GenerateWorker(self.sd, message)
            worker.moveToThread(thread)
            thread.started.connect(worker.run)
            worker.finished.connect(thread.quit)
            worker.finished.connect(worker.deleteLater)
            thread.finished.connect(thread.deleteLater)
            worker.error.connect(self.handle_error)
            self._threads.append(thread)
            self._workers.append(worker)
            thread.start()

    def handle_error(self, error_message):
        print(f"Error: {error_message}")

    def on_interrupt_image_generation_signal(self, _data=None):
        if self.sd:
            self.sd.interrupt_image_generation()

    def on_change_scheduler_signal(self, data: dict):
        if self.sd:
            self.sd.load_scheduler(data["scheduler"])

    def on_model_status_changed_signal(self, message: dict):
        if self.sd and message["model"] == ModelType.SD:
            if self.__requested_action is ModelAction.CLEAR:
                self.on_unload_stablediffusion_signal()
            self.__requested_action = ModelAction.NONE



import torch


def get_torch_device(card_index: int = 0):
    use_cuda = torch.cuda.is_available()
    if not use_cuda:
        print("WARNING: CUDA NOT AVAILABLE, USING CPU")
    return torch.device(f"cuda:{card_index}" if use_cuda else "cpu")



from PySide6.QtCore import Slot
from airunner.windows.setup_wizard.base_wizard import BaseWizard


class AgreementPage(BaseWizard):
    setting_key = ""

    def __init__(self, *args):
        super(AgreementPage, self).__init__(*args)
        self.agreed = False

    @Slot(bool)
    def agreement_clicked(self, val):
        self.agreed = val
        self.update_application_settings(self.setting_key, val)




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'about.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QDialog, QDialogButtonBox,
    QFormLayout, QLabel, QSizePolicy, QWidget)

class Ui_about_window(object):
    def setupUi(self, about_window):
        if not about_window.objectName():
            about_window.setObjectName(u"about_window")
        about_window.resize(270, 150)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(about_window.sizePolicy().hasHeightForWidth())
        about_window.setSizePolicy(sizePolicy)
        about_window.setMinimumSize(QSize(270, 150))
        about_window.setMaximumSize(QSize(270, 150))
        self.layoutWidget = QWidget(about_window)
        self.layoutWidget.setObjectName(u"layoutWidget")
        self.layoutWidget.setGeometry(QRect(10, 10, 249, 134))
        self.formLayout = QFormLayout(self.layoutWidget)
        self.formLayout.setObjectName(u"formLayout")
        self.formLayout.setContentsMargins(0, 0, 0, 0)
        self.title = QLabel(self.layoutWidget)
        self.title.setObjectName(u"title")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.title.sizePolicy().hasHeightForWidth())
        self.title.setSizePolicy(sizePolicy1)
        font = QFont()
        font.setPointSize(14)
        self.title.setFont(font)

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.title)

        self.label_3 = QLabel(self.layoutWidget)
        self.label_3.setObjectName(u"label_3")
        sizePolicy1.setHeightForWidth(self.label_3.sizePolicy().hasHeightForWidth())
        self.label_3.setSizePolicy(sizePolicy1)
        self.label_3.setMinimumSize(QSize(0, 0))
        font1 = QFont()
        font1.setPointSize(11)
        self.label_3.setFont(font1)

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.label_3)

        self.label_2 = QLabel(self.layoutWidget)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setFont(font1)

        self.formLayout.setWidget(2, QFormLayout.LabelRole, self.label_2)

        self.buttonBox = QDialogButtonBox(self.layoutWidget)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Ok)

        self.formLayout.setWidget(3, QFormLayout.LabelRole, self.buttonBox)


        self.retranslateUi(about_window)
        self.buttonBox.accepted.connect(about_window.accept)
        self.buttonBox.rejected.connect(about_window.reject)

        QMetaObject.connectSlotsByName(about_window)
    # setupUi

    def retranslateUi(self, about_window):
        about_window.setWindowTitle(QCoreApplication.translate("about_window", u"About", None))
        self.title.setText(QCoreApplication.translate("about_window", u"AI Runner v", None))
        self.label_3.setText(QCoreApplication.translate("about_window", u"Powered by open source software", None))
        self.label_2.setText(QCoreApplication.translate("about_window", u"Copyright \u00a9 2022-2023 Capsize LLC", None))
    # retranslateUi




from PySide6.QtCore import Slot

from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.model_manager.templates.model_ui import Ui_model_widget


class ModelWidget(BaseWidget):
    widget_class_ = Ui_model_widget

    def __init__(self, *args, **kwargs):
        self.icons = [
            ("eye-look-icon", "toolButton"),
            ("recycle-bin-line-icon", "delete_button"),
        ]
        self.path = kwargs.pop("path")
        self.branch = kwargs.pop("branch")
        self.version = kwargs.pop("version")
        self.category = kwargs.pop("category")
        self.pipeline_action = kwargs.pop("pipeline_action")
        self.pipeline_class = kwargs.pop("pipeline_class")
        self.prompts = kwargs.pop("prompts", [])
        super().__init__(*args, **kwargs)
        self.hide_details()

        # details is a QTableWidget
        # we want to set the text of the second column, first row:
        # first, we will get the current item in the table
        self.ui.name.setText(self.path)
        self.ui.details.item(0, 1).setText(self.path)
        self.ui.details.item(1, 1).setText(self.branch)
        self.ui.details.item(2, 1).setText(self.version)
        self.ui.details.item(3, 1).setText(self.category)
        self.ui.details.item(4, 1).setText(self.pipeline_action)
        self.ui.details.item(5, 1).setText(self.pipeline_class)
        self.ui.details.item(6, 1).setText(",".join(self.prompts))

    @Slot(int, int)
    def action_cell_changed(self, row: int, column: int):
        # TODO: Implement this method
        pass

    def hide_details(self):
        # self.ui.details is a QGridLayout
        self.ui.details.hide()

    def show_details(self):
        self.ui.details.show()

    def action_toggled_button_details(self, val):
        self.show_details() if val else self.hide_details()

    def action_clicked_button_settings(self):
        print("edit button clicked")

    def action_clicked_button_delete(self):
        print("delete button clicked")



imagefilter_bootstrap_data = {
    'pixel_art': {
        'display_name': 'Pixel Art',
        'auto_apply': False,
        'name': 'pixel_art',
        'filter_class': 'PixelFilter',
        'image_filter_values': {
            'number_of_colors': {
                'name': 'number_of_colors',
                'value': '24',
                'value_type': 'int',
                'min_value': '2',
                'max_value': '1024'
            },
            'smoothing': {
                'name': 'smoothing',
                'value': '0',
                'value_type': 'int',
                'min_value': '0',
                'max_value': '100'
            },
            'base_size': {
                'name': 'base_size',
                'value': '256',
                'value_type': 'int',
                'min_value': '2',
                'max_value': '256'
            },
        }
    },
    'gaussian_blur': {
        'display_name': 'Gaussian Blur',
        'auto_apply': False,
        'name': 'gaussian_blur',
        'filter_class': 'GaussianBlur',
        'image_filter_values': {
            'radius': {
                'name': 'radius',
                'value': '0.0',
                'value_type': 'float',
                'min_value': None,
                'max_value': None
            }
        }
    },
    'box_blur': {
        'display_name': 'Box Blur',
        'auto_apply': False,
        'name': 'box_blur',
        'filter_class': 'BoxBlur',
        'image_filter_values': {
            'radius': {
                'name': 'radius',
                'value': '0.0',
                'value_type': 'float',
                'min_value': None,
                'max_value': None
            }
        }
    },
    'film': {
        'display_name': 'Film',
        'auto_apply': False,
        'name': 'film',
        'filter_class': 'FilmFilter',
        'image_filter_values': {
            'radius': {
                'name': 'radius',
                'value': '0.0',
                'value_type': 'float',
                'min_value': None,
                'max_value': None
            },
            'red': {
                'name': 'red',
                'value': '0.0',
                'value_type': 'float',
                'min_value': None,
                'max_value': None
            },
            'green': {
                'name': 'green',
                'value': '0.0',
                'value_type': 'float',
                'min_value': None,
                'max_value': None
            },
            'blue': {
                'name': 'blue',
                'value': '0.0',
                'value_type': 'float',
                'min_value': None,
                'max_value': None
            },
        }
    },
    'color_balance': {
        'display_name': 'Color Balance',
        'auto_apply': False,
        'name': 'color_balance',
        'filter_class': 'ColorBalanceFilter',
        'image_filter_values': {
            'cyan_red': {
                'name': 'cyan_red',
                'value': '0.0',
                'value_type': 'float',
                'min_value': None,
                'max_value': None
            },
            'magenta_green': {
                'name': 'magenta_green',
                'value': '0.0',
                'value_type': 'float',
                'min_value': None,
                'max_value': None
            },
            'yellow_blue': {
                'name': 'yellow_blue',
                'value': '0.0',
                'value_type': 'float',
                'min_value': None,
                'max_value': None
            }
        }
    },
    'halftone': {
        'display_name': 'Halftone Filter',
        'auto_apply': False,
        'name': 'halftone',
        'filter_class': 'HalftoneFilter',
        'image_filter_values': {
            'sample': {
                'name': 'sample',
                'value': '1',
                'value_type': 'int',
                'min_value': None,
                'max_value': None
            },
            'scale': {
                'name': 'scale',
                'value': '1',
                'value_type': 'int',
                'min_value': None,
                'max_value': None
            },
            'color_mode': {
                'name': 'color_mode',
                'value': 'L',
                'value_type': 'str',
                'min_value': None,
                'max_value': None
            },
        }
    },
    'registration_error': {
        'display_name': 'Registration Error',
        'auto_apply': False,
        'name': 'registration_error',
        'filter_class': 'RegistrationErrorFilter',
        'image_filter_values': {
            'red_offset_x_amount': {
                'name': 'red_offset_x_amount',
                'value': '3',
                'value_type': 'int',
                'min_value': None,
                'max_value': None
            },
            'red_offset_y_amount': {
                'name': 'red_offset_y_amount',
                'value': '3',
                'value_type': 'int',
                'min_value': None,
                'max_value': None
            },
            'green_offset_x_amount': {
                'name': 'green_offset_x_amount',
                'value': '6',
                'value_type': 'int',
                'min_value': None,
                'max_value': None
            },
            'green_offset_y_amount': {
                'name': 'green_offset_y_amount',
                'value': '6',
                'value_type': 'int',
                'min_value': None,
                'max_value': None
            },
            'blue_offset_x_amount': {
                'name': 'blue_offset_x_amount',
                'value': '9',
                'value_type': 'int',
                'min_value': None,
                'max_value': None
            },
            'blue_offset_y_amount': {
                'name': 'blue_offset_y_amount',
                'value': '9',
                'value_type': 'int',
                'min_value': None,
                'max_value': None
            },
        }
    },
    'unsharp_mask': {
        'display_name': 'Unsharp Mask',
        'auto_apply': False,
        'name': 'unsharp_mask',
        'filter_class': 'UnsharpMask',
        'image_filter_values': {
            'radius': {
                'name': 'radius',
                'value': '0.5',
                'value_type': 'float',
                'min_value': None,
                'max_value': None
            },
            'percent': {
                'name': 'percent',
                'value': '0.5',
                'value_type': 'float',
                'min_value': None,
                'max_value': None
            },
            'threshold': {
                'name': 'threshold',
                'value': '0.5',
                'value_type': 'float',
                'min_value': None,
                'max_value': None
            },
        }
    },
    'saturation': {
        'display_name': 'Saturation Filter',
        'auto_apply': False,
        'name': 'saturation',
        'filter_class': 'SaturationFilter',
        'image_filter_values': {
            'factor': {
                'name': 'factor',
                'value': '1.0',
                'value_type': 'float',
                'min_value': None,
                'max_value': None
            },
        }
    },
    'invert': {
        'display_name': 'Invert',
        'auto_apply': False,
        'name': 'invert',
        'filter_class': 'Invert',
        'image_filter_values': {}
    },
    'rgb_noise': {
        'display_name': 'RGB Noise Filter',
        'auto_apply': False,
        'name': 'rgb_noise',
        'filter_class': 'RGBNoiseFilter',
        'image_filter_values': {
            'red': {
                'name': 'red',
                'value': '0.0',
                'value_type': 'float',
                'min_value': None,
                'max_value': None
            },
            'green': {
                'name': 'green',
                'value': '0.0',
                'value_type': 'float',
                'min_value': None,
                'max_value': None
            },
            'blue': {
                'name': 'blue',
                'value': '0.0',
                'value_type': 'float',
                'min_value': None,
                'max_value': None
            },
        }
    }
}



from PySide6.QtWidgets import QWizard
from airunner.mediator_mixin import MediatorMixin
from airunner.utils.os.create_airunner_directory import create_airunner_paths
from airunner.windows.main.settings_mixin import SettingsMixin
from airunner.windows.setup_wizard.installation_settings.install_failed_page import InstallFailedPage
from airunner.windows.setup_wizard.installation_settings.install_page import InstallPage
from airunner.windows.setup_wizard.installation_settings.install_success_page import InstallSuccessPage
from airunner.windows.setup_wizard.path_settings.path_settings import PathSettings


class DownloadWizardWindow(QWizard, MediatorMixin, SettingsMixin):
    """
    The download wizard window class for AI Runner.
    This class is used to download models and other resources required for AI Runner.
    """
    def __init__(self):
        """
        Initialize the download wizard window.
        :param setup_settings: The setup settings dictionary.
        """
        MediatorMixin.__init__(self)
        
        super(DownloadWizardWindow, self).__init__()
        self.setWindowTitle("AI Runner Download Wizard")
        self.setWizardStyle(QWizard.WizardStyle.ModernStyle)
        self.setOption(QWizard.WizardOption.IndependentPages, True)

        self.button(
            QWizard.WizardButton.FinishButton
        ).clicked.connect(self.save_settings)

        self.button(QWizard.WizardButton.NextButton).clicked.connect(self.next_button_clicked)

        self.init_pages()

    def save_settings(self):
        """
        Override this function to save settings based on specific page in question.
        Do not call this function directly.
        :return:
        """
        self.update_application_settings("run_setup_wizard", False)
        self.update_application_settings("download_wizard_completed", True)

    def next_button_clicked(self):
        current_page = self.currentPage()
        print("NEXT BUTTON CLICKED", current_page)
        if hasattr(current_page, 'next'):
            print("PAGE HAS NEXT")
            current_page.next()
        if hasattr(current_page, "start"):
            print("PAGE HAS RUN")
            current_page.start()

    def init_pages(self):
        """
        Initialize the wizard pages based on setup settings.
        """
        failed = True

        if (
            self.application_settings.user_agreement_checked and
            self.application_settings.stable_diffusion_agreement_checked and
            self.application_settings.airunner_agreement_checked
        ):
            self.setPage(0, PathSettings(self))
            self.setPage(1, InstallPage(self))
            self.setPage(2, InstallSuccessPage(self))
            failed = False

        if failed:
            self.setPage(2, InstallFailedPage(self))

    def show_final_page(self):
        """
        Show the final page.
        """
        # Find the ID of the InstallSuccessPage
        for page_id in self.pageIds():
            page = self.page(page_id)
            if isinstance(page, InstallSuccessPage):
                self.setCurrentId(page_id)
                return



from PIL.ImageFilter import UnsharpMask as PILUnsharpMask
from airunner.filters.base_filter import BaseFilter


class UnsharpMask(BaseFilter):
    def apply_filter(self, image, do_reset):
        return image.filter(PILUnsharpMask(radius=self.radius * 100, percent=int(self.percent * 200), threshold=int(self.threshold * 10)))
